/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"


int incoming[1600][1600]; int incoming_used[1600];
int outgoing[1600][1600]; int outgoing_used[1600];


/******************************************************************************
* GET_TRANSITIONS returns the transition matrix, a states list and states
* maps.
*   [big_mat list maps] = GET_TRANSITIONS(freq, comb) computes and returns
*   the transition matrix (1600x1600), a list representing the first states
*   of the first allele in letterspace (so a 40-vector , since the a
*   complete states is made of two alleles => 40x40 = 1600-vector), and
*   some more lists of the states in the maps structure.
*
*   N.B.: a capital letter represents a gap state with the memory of that
*   letter. For example, an 'A' represents the gap '-' with the memory that
*   the last letter seen is an a. So 'A' means '-_a', etc.
*
*   Parameters:
*   freq - a struct holding the fields:
*       freq.a - the frequency of 'a' nucleotides (e.g. 0.25)
*       freq.c - the frequency of 'c' nucleotides (e.g. 0.25)
*       freq.g - the frequency of 'g' nucleotides (e.g. 0.25)
*       freq.t - the frequency of 't' nucleotides (e.g. 0.25)
*       freq.gapopen - the penalty for gap open
*       freq.gapextend - the penalty for gap extend
*   comb - a struct holding the fields:
*       comb.c - a 5x8 matrix representing the 8 states where each of the 5
*           colors (0-3 and gap[4]) are present
*       comb.l - a 5x8 matrix representing the 8 states where each of the 5
*           possible letters (0-3 = {a,c,g,t} and gap[4]) are present
*
*   Returns:
*   big_mat - the 1600x1600 transition probability matrix
*   list - the 40x2 list of the first allele letters (e.g. {aa, ac, ...})
*   maps - a struct holding states to letter or states to color maps.
*       maps.full_letter - 1600x4 matrix, the letters of each of the 1600
*           states for both alleles. For example, the second state will be
*           aa/ac, which will be represented in ns as [0 0 0 1]
*       maps.call_letter - 1600x2 matrix, only the call letter of each of
*           the states for both alleles. For example, the second state will
*           be aa/ac, with the call letters a/c, which will be represented
*           in ns as [0 1]
*       maps.call_color - 1600x2 matrix, reprenting the call colors of each
*           of the states for both alleles. For example, the second state
*           will be aa/ac, which is 0/1 in colorspace, which will be
*           represented as [0 1]
*
*   Author: Adrian Dalca
*   TODO: not fully tested the final big_mat result...
*   NOTE: there currently isn't any snp penalty implemented in the
*       resulting transition matrix...
*******************************************************************************/
void get_transitions(
    Comb_t         comb,
    VA_MatDouble * big_mat,
    Maps_t       * maps)
{
    VA_UINT32 yl;
    VA_UINT32 xl;
    VA_UINT32 x1;
    VA_UINT32 x2;
    VA_UINT32 y1;
    VA_UINT32 y2;

    char * letters = "acgt";
    char * gaps = "ACGT";
    int sz = sizeof(state_list) / sizeof(state_list[0]);
    int i;
    VA_MatDouble trans_mat = va_allocMatDouble(sz, sz,"trans_mat",__FILE__,__LINE__);

#ifndef NDEBUG
    VA_ASSERT(big_mat != NULL);
    VA_ASSERT(maps != NULL);
#endif


    memset(incoming_used,0,1600*sizeof(int));
    memset(outgoing_used,0,1600*sizeof(int));
    /* build first (primary) transition matrix (40 to 40 states) */

    /* set-up transition matrices*/
    //memset(trans_mat.data, 0, trans_mat.size_x * 
    //    trans_mat.size_y * sizeof(VA_DOUBLE));

    /* go through each element in the lists */
    for (yl = 0; yl < sz; yl++) {
        char * y = state_list[yl];

        for (xl = 0; xl < sz; xl++) {
            char * x = state_list[xl];

            /* so now we have two vectors of length 2: the first state
             * y(1:2) and the second state (the one we are transitiong TO)
             * x(1:2)
             */

	    //trans_mat.data[yl * sz + xl] = -log(25);
	    trans_mat.data[yl * sz + xl] =0.0;
            /* if inner letters match and they are nucleotides (not gaps) */
	    if ((y[1] == x[0]) && (strchr(letters, y[1]) != NULL)) {
                /*if (strchr(gaps, x[1]) != NULL) {
                    trans_mat.data[yl * sz + xl] = insert_open;
                } else {
                    switch (x[1]) {
                        case 'a':
                            trans_mat.data[yl * sz + xl] = freq_a;
                            break;
                        case 'c':
                            trans_mat.data[yl * sz + xl] = freq_c;
                            break;
                        case 'g':
                            trans_mat.data[yl * sz + xl] = freq_g;
                            break;
                        case 't':
                            trans_mat.data[yl * sz + xl] = freq_t;
                            break;
                    } 
                }*/
            } else if (y[1] == x[0] && (strchr(gaps, y[1]) != NULL)) {
                /*if (strchr(gaps, x[1]) != NULL) {
                    trans_mat.data[yl * sz + xl] = insert_extend;
                } else {    
                    switch (x[1]) {
                        case 'a':
                            trans_mat.data[yl * sz + xl] = freq_a;
                            break;
                        case 'c':
                            trans_mat.data[yl * sz + xl] = freq_c;
                            break;
                        case 'g':
                            trans_mat.data[yl * sz + xl] = freq_g;
                            break;
                        case 't':
                            trans_mat.data[yl * sz + xl] = freq_t;
                            break;
                    } 
                }*/
            } else {
		trans_mat.data[yl * sz + xl] = -1.0/0.0;
	    }
        }
    }
    /* secondary transition matrix (40x40 to 40x40 states) */
      
    /* prepare the final transition mat. */
    *big_mat = va_allocMatDouble(sz * sz, sz * sz, "big_mat",__FILE__,__LINE__);
    /* state transitioning FROM: */
    /* the list index (1-40) of the first allele */
    for (y1 = 0; y1 < sz; y1++)
    {
        /* the list index (1-40) of the second allele */ 
        for (y2 = 0; y2 < sz; y2++)
        {
            /* state transitioning TO: */
            /* the list index (1-40) of the first allele */
            for (x1 = 0; x1 < sz; x1++)
            {
                /* the list index (1-40) of the second allele */
                for (x2 = 0; x2 < sz; x2++)
                {
		    int x = x1 * sz + x2; int y = y1 * sz + y2;
		    if (haploid==1 && (y1!=y2 || x1!=x2)) {
			big_mat->data[y * big_mat->size_x + x] = -1.0/0.0;
		    } else {
                    	big_mat->data[y * big_mat->size_x + x] = 
                        	trans_mat.data[y1 * trans_mat.size_x + x1] +
                        	trans_mat.data[y2 * trans_mat.size_x + x2]; 
		    }
		    //going form x to y
		    if (!isinf(big_mat->data[ y * big_mat->size_x + x])) {
			incoming[x][incoming_used[x]++] = y;
			outgoing[y][outgoing_used[y]++] = x;	
		    }
                }
            }
        }
    }
    free(trans_mat.data);
    int x; int y;
    for (y=0; y<1600; y++) {
	int used=outgoing_used[y];
	//fprintf(stderr,"%d\n",used);
	if (used>0) {
		double log_value = -log(used);
		for (x=0; x<used; x++) {
			int to_state = outgoing[y][x];
			/*fprintf(stderr,"%c%c/%c%c -> %c%c/%c%c , %d\n",
				state_list[y%40][0],state_list[y%40][1],
				state_list[y/40][0],state_list[y/40][1],
				state_list[to_state%40][0],state_list[to_state%40][1],
				state_list[to_state/40][0],state_list[to_state/40][1],
				used);*/
			big_mat->data[y * big_mat->size_x + to_state]+=log_value;	
		}
	} 
    }
	//exit(1);

 
    /* build several state maps */
    
    /* full_letter has states such as 'ac/cg' i.e. the letterspace state 
     * 'ac/cg' would be given as [0 1 1 2]
     */
    maps->full_letter = va_allocMatU8(4, sz * sz);
    memset(maps->full_letter.data, 0, 
        maps->full_letter.size_x * maps->full_letter.size_y);

    for (i = 0; i < sz; i++)
    {
        int j;

        for (j = 0; j < sz; j++)
        {
            char ref[4];
            int k;
            int idx;
            VA_VecU8 r;
            VA_VecU8 p;

            ref[0] = state_list[i][0];
            ref[1] = state_list[i][1];
            ref[2] = state_list[j][0];
            ref[3] = state_list[j][1];

            r.data = (VA_UINT8*)ref;
            r.size = sizeof(ref);
 
            p = nuc2nuc(r);
#ifndef NDEBUG
            VA_ASSERT(p.data != NULL);
#endif

            for (k = 0; k < 4; k++)
            {
                idx = (i * sz + j) * maps->full_letter.size_x + k;
                maps->full_letter.data[idx] = p.data[k];
            }
            free(p.data);
       }
    }

    //find out what are the various lengths
    /*int lengths[1600]; int used=0;
    for (y1=0; y1<1600; y1++) {
	int i; int found=0; int looking_for=incoming_used[y1];
	for (i=0; i<used; i++) {
		if (lengths[i]==looking_for) {
			found=1;
		}
	}
	if (found==0) {
		VA_UINT8 temp[4]; 
		temp[0]=maps->full_letter.data[ y1*4 + 0];
		temp[1]=maps->full_letter.data[ y1*4 + 1];
		temp[2]=maps->full_letter.data[ y1*4 + 2];
		temp[3]=maps->full_letter.data[ y1*4 + 3];
		VA_VecU8 temp_vec; temp_vec.size=4; temp_vec.data=temp;
		temp_vec = nuc2nuc(temp_vec);
		printf("%c%c/%c%c\n",temp_vec.data[0],temp_vec.data[1],temp_vec.data[2],temp_vec.data[3]);
		lengths[used++]=looking_for;
	}
    }
    for (y1=0; y1<used; y1++) {
	printf("LENGTH %d\n",lengths[y1]);
    }
    exit(1);*/ 
    /* just the potential output letters in the state */

    maps->call_letter = va_allocMatU8(2, sz * sz);
    memset(maps->call_letter.data, 0, maps->call_letter.size_x *
        maps->call_letter.size_y);

    for (i = 0; i < sz; i++)
    {
        int j;

        for (j = 0; j < sz; j++)
        {
            int idx1 = (i * sz + j) * maps->call_letter.size_x;
            int idx2 = (i * sz + j) * maps->full_letter.size_x;

            maps->call_letter.data[idx1] = maps->full_letter.data[idx2 + 1];
            maps->call_letter.data[idx1 + 1] = maps->full_letter.data[idx2 + 3];
        }
    }

    /* the colors of the states */
    maps->call_color = va_allocMatU8(2, sz * sz);
    memset(maps->call_color.data, -1, maps->call_color.size_x *
        maps->call_color.size_y);
    {
        int color_1;
        size_t csz0 = sizeof(comb.c[0])/sizeof(comb.c[0][0]);
        size_t csz1 = sizeof(comb.c)/sizeof(comb.c[0]);
     
        for (color_1 = 0; color_1 < csz1; color_1++)
        {
            int pos_1;
            for (pos_1 = 0; pos_1 < csz0; pos_1++)
            {
                int color_2;
                for (color_2 = 0; color_2 < csz1; color_2++)
                {
                    int pos_2;
                    for (pos_2 = 0; pos_2 < csz0; pos_2++)
                    {    
                        int pos = (comb.c[color_1][pos_1] - 1) * sz + 
                                  comb.c[color_2][pos_2] - 1;
                        maps->call_color.data[pos * maps->call_color.size_x] = color_1 + 1;
                        maps->call_color.data[pos * maps->call_color.size_x + 1] = color_2 + 1;
                    }
                }
            }
        }
#ifndef NDEBUG 
        for (i = 0; i < maps->call_color.size_x * maps->call_color.size_y; i++)
            VA_ASSERT(maps->call_color.data[i] != 0);
#endif
    }


    // addition in may/june 2010 - a map of the symmetric states.
    maps->symm = va_allocVecU32(sz * sz);
    unsigned int state, i1, i2;
    for (state = 0; state < sz * sz; state++) {
      maps->symm.data[state] = 1601;
    }

    for (y1 = 0; y1 < sz; y1++) { // first allele
      for (y2 = 0; y2 < sz; y2++) { // second allele

	i1 = y1 * sz + y2;
	i2 = y2 * sz + y1;
	maps->symm.data[i1] = i2;

      } // y1
    } // y2
    
    //rc = va_GetTime(&te);
    //VA_ASSERT(rc == 0);

    /*for (i=0; i<1600; i++) {
	double sum = 0.0;
	int j; 
	for (j=0; j<1600; j++) {
		sum+=big_mat->data[j*1600 + i];
	}
	VA_VecU8 temp; temp.size=4; temp.data=maps->full_letter.data + i*4; temp=nuc2nuc(temp);
	VA_UINT8 * d = temp.data;
	printf("%5d %5.1e %c%c/%c%c\n",i,sum,d[0],d[1],d[2],d[3]);
	free(d);
    }*/


    //tdiff = va_TimeSub(&te, &ts);
    //fprintf(stderr, "\t Done in: %ld.%03ld secs\n",
     //   tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);
}    

