/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

typedef struct VA_ValIndex VA_ValIndex;
struct VA_ValIndex {
    VA_DOUBLE value;
    VA_UINT32 idx;
};

static int compareRec(
    const void * rec1,
    const void * rec2);

/******************************************************************************
* PREDICT offers a first-level of prediction of the 2 haplomes 
*   [maxvals_2, lmat, cmat] = PREDICT(p, maps) offers a first-level of
*   prediction of the 2 haplomes given the probability matrix and the
*   letter/color maps.
*
*   Parameters
*   p - the probability matrix, dim #states x length-of-genome.
*       so with the most recent model, 1600 x gen_len
*   maps - a struct holding states to letter or states to color maps.
*       maps.full_letter - 1600x4 matrix, the letters of each of the 1600
*           states for both alleles. For example, the second state will be
*           aa/ac, which will be represented in ns as [0 0 0 1]
*       maps.call_letter - 1600x2 matrix, only the call letter of each of
*           the states for both alleles. For example, the second state will
*           be aa/ac, with the call letters a/c, which will be represented
*           in ns as [0 1]
*       maps.call_color - 1600x2 matrix, reprenting the call colors of each
*           of the states for both alleles. For example, the second state
*           will be aa/ac, which is 0/1 in colorspace, which will be
*           represented as [0 1]
*
*   Returns
*       maxvals_2 - the probability of the maximum states and the 
*           second-maximum states (so a 2 x gen_len matrix) in the 25-state
*       lmat - the letter matrix for the two alleles. 
*       cmat - the color matrix for the two allels (a translation of lmat)
*
*   Author: Adrian V. Dalca
*******************************************************************************/
void predict(
    VA_MatDouble   p, 
    Maps_t         maps,
    VA_MatDouble * maxvals_2,
    VA_MatU8     * lmat,
    VA_MatU8     * cmat)
{
    VA_MatDouble  p2; 
    int i, j, k;
    VA_ValIndex * p2_sort;
#ifndef NDEBUG
    VA_ASSERT(maxvals_2 != NULL);
    VA_ASSERT(lmat != NULL);
    VA_ASSERT(cmat != NULL);
#endif


    


    /* states are symmetric. add symmetric states
    p2 = va_allocMatDouble(p.size_x, p.size_y);
    memset(p2.data, 0, p2.size_x * p2.size_y * sizeof(VA_DOUBLE));


    for (i = 0; i < 1600; i++) {

      uj = maps.symm.data[i];
      VA_ASSERT(maps.symm.data[uj] == i);
      VA_ASSERT(uj >= 0 && uj <= 1599);

      
      if (uj == i) {
	for (k = 0; k < p2.size_x; k++) {
	  p2.data[i * p2.size_x + k] = p.data[i * p.size_x + k];
	}

      } else if (i > uj) {

	for (k = 0; k < p2.size_x; k++) {
	  v1 = p.data[i * p.size_x + k];
	  v2 = p.data[uj * p.size_x + k];

	  if (v1 >= v2) {
	    p2.data[i * p2.size_x + k] = v1 + v2;
	    
	  } else {
	    p2.data[uj * p2.size_x + k] = v1 + v2;

	  } 

	}
      }
    }*/
    
    

    /* add votes for the same alleles together */ 
    p2 = va_allocMatDouble(p.size_x, p.size_y,"p2",__FILE__,__LINE__);
    memset(p2.data, 0, p2.size_x * p2.size_y * sizeof(VA_DOUBLE));

    int let1, clet1;
    int let2, clet2;

    int used_mat[5][5];
    for (i = 0; i < 5; i++)
      for (j = 0; j < 5; j++)
	used_mat[i][j] = 0;
   
   

    for (i = 0; i < 1600; i++) {

      let1 = maps.call_letter.data[i * 2];
      let2 = maps.call_letter.data[i * 2 + 1];

      // if not yet dealt with this allel pair
      if (used_mat[let1][let2] == 0 && used_mat[let2][let1] == 0) {
	
	for (k = 0; k < p2.size_x; k++)  
	  p2.data[i * p2.size_x + k] = p.data[k*1600  + i];

	// go through all future states and add that allele pair
	for (j = i + 1; j < 1600; j++) {
	  
	  clet1 = maps.call_letter.data[j * 2];
	  clet2 = maps.call_letter.data[j * 2 + 1];
	  
	  if ((let1 == clet1 && let2 == clet2) || (let1 == clet2 && let2 == clet1))
	    for (k = 0; k < p2.size_x; k++)  
	      p2.data[i * p2.size_x + k] += p.data[k*1600 + j];
	  
	  
	}
	used_mat[let1][let2] = 1;
	used_mat[let2][let1] = 1;
      }

    }



    /* get the maximum values */
    p2_sort = (VA_ValIndex *)malloc(p2.size_y * p2.size_x * sizeof(VA_ValIndex));
#ifndef NDEBUG
    VA_ASSERT(p2_sort != NULL);
#endif

    /* get maximum states, and the second-max states. */
    for (i = 0; i < p2.size_x; i++)
    {
        for (j = 0; j < p2.size_y; j++)
        {
            p2_sort[i * p2.size_y + j].value =
                p2.data[j * p2.size_x + i];
            p2_sort[i * p2.size_y + j].idx = j + 1;
        }
        qsort(&p2_sort[i * p2.size_y], p2.size_y,
              sizeof(VA_ValIndex), compareRec);
    }

    *maxvals_2 = va_allocMatDouble(p2.size_x, 2,"maxvals_2",__FILE__,__LINE__);
    for (i = 0; i < maxvals_2->size_x; i++)
    {
        maxvals_2->data[i] = p2_sort[p2.size_y * i].value;
        maxvals_2->data[maxvals_2->size_x + i] = p2_sort[p2.size_y * i + 1].value;
    }

    /* get the letter predictions */
    
    *lmat = va_allocMatU8(p.size_x, 2);
    for (i = 0; i < lmat->size_y; i++)
        for (j = 0; j < lmat->size_x; j++)
            lmat->data[i * lmat->size_x + j] = 
                maps.call_letter.data[(p2_sort[j * p2.size_y].idx - 1) * 
                                      maps.call_letter.size_x + i];

    /* 
     * N.B. colors are just a translation of the letters, and NOT taken from
     * the states !!!
     */
    
    free(p2_sort);

    {
        VA_VecU8 lmat1;
        VA_VecU8 lmat2;
        VA_VecU8 lmat1_res;
        VA_VecU8 lmat2_res;

        lmat1 = va_allocVecU8(lmat->size_x);
        memcpy(lmat1.data, lmat->data, lmat1.size * sizeof(VA_UINT8));
        lmat2 = va_allocVecU8(lmat->size_x);
        memcpy(lmat2.data, &lmat->data[lmat->size_x], lmat2.size * sizeof(VA_UINT8));

        lmat1_res = nuc2col(lmat1);
        lmat2_res = nuc2col(lmat2);

        *cmat = va_allocMatU8(lmat->size_x, 2);
        memcpy(&cmat->data[1], lmat1_res.data, 
               sizeof(VA_UINT8) * lmat1_res.size); 
        memcpy(&cmat->data[cmat->size_x + 1], lmat2_res.data, 
               sizeof(VA_UINT8) * lmat2_res.size); 
        cmat->data[0] = 0;
        cmat->data[cmat->size_x] = 0;
	free(lmat1.data);
	free(lmat2.data);
        free(lmat1_res.data);
        free(lmat2_res.data);
    }
    
    /* 
     * should check its the same as a cmat from the states, although it may
     * not be...
     */
    free(p2.data); 
	    /* print time */
    //rc = va_GetTime(&te);
   // VA_ASSERT(rc == 0);

    //tdiff = va_TimeSub(&te, &ts);

    //fprintf(stderr, "\t Done in: %ld.%03ld secs\n",
//	tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}
    
static int compareRec(
    const void * rec1,
    const void * rec2)
{
    VA_ValIndex p1;
    VA_ValIndex p2;
    int ret = 0;
#ifndef NDEBUG
    VA_ASSERT((rec1 != NULL) && (rec2 != NULL));
#endif

    p1 = *(VA_ValIndex *)rec1;
    p2 = *(VA_ValIndex *)rec2;

    if (p2.value > p1.value)
       ret = 1;
    else if (p1.value > p2.value)
       ret = -1;
    else 
       ret = p1.idx - p2.idx;

    return (ret);
}

