/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* VARiD run the VARiD algorithm
*   VARiD(cdata, ref_pos) run the cosin algorithm, where cdata is
*   filled as requested, and ref_pos is a dictionary of extra positions
*   and respective nucleotides to allow in the reference when
*   transitioning. More details are coming
*
*   Quick breakdown:
*       1) compute the transition breakdown, which determines what the
*       probability of transitioning from one state to another is.
*       2) compute the covmap and emission probability. The covmap computes
*       the coverages and some probabilities associates with the reads, and
*       then we compute the emission probability based on these coverages.
*       3) compute the forward and backward algorithm to get a final
*       probability.
*
*
*   TODO: verify code for probability matrix, normalization, and sanity
*   check
******************************************************************************/

static Comb_t comb = {
    {
        {1,   6,  11,  16,  25,  30,  35,  40},    /* color 1 */
        {2,   5,  12,  15,  26,  29,  36,  39},    /* color 2 */
        {3,   9,   8,  14,  27,  33,  32,  38},    /* color 3 */
        {4,   7,  10,  13,  28,  31,  34,  37},    /* color 4 */
        {17, 18,  19,  20,  21,  22,  23,  24}     /* gap */
    },
    {
        {1,   5,   9,  13,  25,  29,  33,  37},    /* letter a */
        {2,   6,  10,  14,  26,  30,  34,  38},    /* letter c */
        {3,   7,  11,  15,  27,  31,  35,  39},    /* letter g */
        {4,   8,  12,  16,  28,  32,  36,  40},    /* letter t */
        {17, 18,  19,  20,  21,  22,  23,  24}     /* no letter emission */
    }
};

void varid(
    VA_INT32       ref_pos,
    Maps_t       * maps,
    Covmaps_t    * covmap,
    VA_MatDouble * p,
    VA_VecU8 * qval_ordering)
{
    int i;
    int j;
    VA_MatDouble trans_mat;
    VA_MatDouble e;
    VA_MatDouble f;
    VA_MatDouble b;
    VA_VecDouble s_f;
    VA_VecDouble s_b;
    VA_VecI32 window;
    //VA_VecU8  skippos;


    
    /* prepare data */
    
    /*
     * we will work with a 40*gen_length matrix (there are 40 possible
     * states for each position). the comb stuct indicates
     * which states each color/letter is present in: for example, we have
     * color 0 in states 1, 6, 11, 16, ...
     *
     * This is because the states are done as follows: (reminder: capital L
     * is a gap with L memory, see get_transitions function)
     *   states 01-16: {a,c,g,t}x{a,c,g,t}={aa, ac, ag, at, ca, ..., tg, tt}
     *   states 17:20: {aA, cC, gG, tT}
     *   states 20:24: {AA, CC, GG, TT}
     *   states 25:40: {A,C,G,T}x{a,c,g,t}={Aa, Ac, Ag, At, Ca, ..., Tg, Tt}
     */
    /* color emissions */
    /* emit color 1:4 (really colors 0:3), and c(5) = gap emission */
#ifndef NDEBUG
    for (i = 0; i < sizeof(comb.c) / sizeof(comb.c[0]); i++)
         for (j = 0; j < sizeof(comb.c[0])/sizeof(comb.c[0][0]); j++)
             VA_ASSERT(comb.c[i][j] != 0);

    for (i = 0; i < sizeof(comb.l) / sizeof(comb.l[0]); i++)
         for (j = 0; j < sizeof(comb.l[0])/sizeof(comb.l[0][0]); j++)
             VA_ASSERT(comb.l[i][j] != 0);
#endif
    
    /* frequencies */
    /* set the frequencies of the various letters in the reference */


    if (get_bp_freq == TRUE ){
	    freq[0] = freq[1] = freq[2] = freq[3] = 0.0; 
	    for (i = 0; i < reference_wgaps->size; i++) {
		freq[0] += (reference_wgaps->number_sequence[i] == 0) ? 1.0 : 0.0; 
		freq[1] += (reference_wgaps->number_sequence[i] == 1) ? 1.0 : 0.0; 
		freq[2] += (reference_wgaps->number_sequence[i] == 2) ? 1.0 : 0.0; 
		freq[3] += (reference_wgaps->number_sequence[i] == 3) ? 1.0 : 0.0; 
	    }
    	    freq[0] /= reference_wgaps->size;
    	    freq[1] /= reference_wgaps->size;
    	    freq[2] /= reference_wgaps->size;
    	    freq[3] /= reference_wgaps->size;
    }

    /*freq[0] -= insert_open/4;
    freq[1] -= insert_open/4;
    freq[2] -= insert_open/4;
    freq[3] -= insert_open/4;

    //take logs?
    freq[0]=log(freq[0]);
    freq[1]=log(freq[1]);
    freq[2]=log(freq[2]);
    freq[3]=log(freq[3]);
    insert_open=log(insert_open);
    insert_extend=log(insert_extend);*/

    /* set the frequenciesof various letters in the reference */

   
    /* 
     * set the gap open and gap extension penalties.
     * TODO: should these be smaller? should they be provided in cdata?
     */

    /* compute the transitions matrices */
    struct timespec time_start; struct timespec time_end; struct timespec time_diff;
    /* computation */
    fprintf(stderr,"   - Computing transistion matrix ...\n"); va_GetTime(&time_start);
    memset(&trans_mat, 0, sizeof(trans_mat));
    get_transitions(comb, &trans_mat, maps);
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);


    /* compute the coverage maps */
    int unique_qvals=0;
    for (i=0; i<qval_ordering->size; i++) {
	unique_qvals=MAX(unique_qvals,qval_ordering->data[i]+1);
    }

    VA_MatI32 letter_qvals = va_allocMatI32(reference_wgaps->size,5*unique_qvals,"letter_qvals", __FILE__ ,__LINE__  );
    memset(letter_qvals.data,0,sizeof(VA_INT32)*5*unique_qvals*reference_wgaps->size);

    VA_MatI32 color_qvals = va_allocMatI32(reference_wgaps->size,5*unique_qvals,"color_qvals", __FILE__ ,__LINE__ );
    memset(color_qvals.data,0,sizeof(VA_INT32)*5*unique_qvals*reference_wgaps->size);

    fprintf(stderr,"   - Computing coverage map ...\n"); va_GetTime(&time_start);
    *covmap = varid_covmaps(&letter_qvals,&color_qvals,qval_ordering,unique_qvals); 
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
   
    /* compute emissions */
    fprintf(stderr,"   - Computing emissions matrix ...\n"); va_GetTime(&time_start);
    e = varid_emissions(*covmap, &letter_qvals, &color_qvals, qval_ordering, unique_qvals);
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    free(letter_qvals.data);
    free(color_qvals.data);
    free(qval_ordering->data);   
 
    /* run forward and backward algorithm */
    window = va_allocVecI32(2);
    window.data[0] = 1;
    window.data[1] = reference_wgaps->size;

    *p = va_allocMatDouble(reference_wgaps->size, trans_mat.size_y,"p",__FILE__,__LINE__);
    memset(p->data, 0, sizeof(VA_DOUBLE) * p->size_x * p->size_y);
    if (viterbi==0){	 

	fprintf(stderr,"   - Computing forward matrix ...\n");
	va_GetTime(&time_start);
    	varid_forward(trans_mat, e, window, &f, &s_f);
	va_GetTime(&time_end);
	time_diff = va_TimeSub(&time_end, &time_start);
	fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);

	fprintf(stderr,"   - Computing backward matrix ...\n");
	va_GetTime(&time_start);
   	varid_backward(trans_mat, e, window, &b,&s_b);
	va_GetTime(&time_end);
	time_diff = va_TimeSub(&time_end, &time_start);
	fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
	double forward_scaling=0.0; double backward_scaling=0.0;
	double sum_forward_scaling=0.0; double sum_backward_scaling=0.0;
	for (i=0; i<s_b.size; i++) {
		sum_backward_scaling+=s_b.data[i];
		sum_forward_scaling+=s_f.data[i];
	} 
	double normalization_constant=0.0;
	for (i=0; i<1600; i++) {
		normalization_constant+=exp(f.data[ (reference_wgaps->size - 1)*1600  + i]); 
	}
	double log_normalization_constant=log(normalization_constant) + sum_forward_scaling;
	for (i=0; i<reference_wgaps->size; i++) {
		forward_scaling+=s_f.data[i];
		//do stuff
		double sum=0.0;
		int j;
		for (j=0; j<1600; j++) {
			int coordinate=i*1600 + j;
			p->data[coordinate] = exp(f.data[coordinate] + b.data[coordinate] - log_normalization_constant + forward_scaling + (sum_backward_scaling - backward_scaling));
			if (isnan(p->data[coordinate])) {
				fprintf(stderr,"varid : possible fatal error, nan appeared in final probabilities!\n");
			}
			sum += p->data[coordinate];
		}
		//fprintf(stderr,"%d %e\n",i,sum);
		//done stuff
		backward_scaling+=s_b.data[i];
	}
        /*VA_VecU8 ls = nuc2nuc(cdata.ref);
	FILE * z;
        z = fopen("f_matrix.txt","w");
        //print_genome(z,cdata.ref);
        print_matrix_double(z,f.data,f.size_x,f.size_y,ls.data,maps->full_letter.data);
        fclose(z);
        z = fopen("b_matrix.txt","w");
        //print_genome(z,cdata.ref);
        print_matrix_double(z,b.data,b.size_x,b.size_y,ls.data,maps->full_letter.data);
        fclose(z);*/
	free(f.data);
	free(s_f.data);
	free(b.data);
	free(s_b.data);
    } else {
	fprintf(stderr,"   - Computing viterbi backtrace matrix\n");
	va_GetTime(&time_start);
    	varid_viterbi(trans_mat, e, window, p);
	va_GetTime(&time_end);
	time_diff = va_TimeSub(&time_end, &time_start);
	fprintf(stderr, "   + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    }

        /*FILE * z = fopen("e_matrix.txt","w");
        //print_genome(z,cdata.ref);
        print_matrix_double(z,e.data,e.size_x,e.size_y,reference_wgaps->sequence,maps->full_letter.data);
        fclose(z);
        z = fopen("p_matrix.txt","w");
        //print_genome(z,cdata.ref);
        print_matrix_double(z,p->data,p->size_x,p->size_y,reference_wgaps->sequence,maps->full_letter.data);
        fclose(z);*/


    free(window.data);
    free(e.data);
    free(trans_mat.data);
 
    /* print time results */
    //rc = va_GetTime(&te);
    //VA_ASSERT(rc == 0);

    //tdiff = va_TimeSub(&te, &ts);

    /* done read prep */
    //fprintf(stderr, "\t DONE VARiD Algorithm & prediction in: %ld.%03ld secs\n",
      //      tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}    
