/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* VARID_COVMAPS computes the letter and colorspace coverage maps
*   Here we compute a few coverage-related stats. We compute the overall
*   coverage, the number of times we see each color and each letter, and
*   finally the emission support for a color (col_em) and against a
*   color (noncol_em), and the same for letters. More details to come.
*
*
*
*
*   GEN_SIZE, READSTARTS, READS
*
*
*   note: stats CUTOFFs, size CUTOFFs
* TODO: quality values
* !!! Asuuming T start when we do first letter translation???
*******************************************************************************/

extern double*  cs_error_cnt;
extern double*  cs_error_dpts;
extern unsigned int cs_error_size;
extern double*  ls_error_cnt;
extern double*  ls_error_dpts;
extern unsigned int ls_error_size;


static int update_map_with_read(VA_VecU8 seq, VA_VecU8 qvals,
				VA_UINT32 start, VA_DOUBLE * qval_errors,
				VA_DOUBLE * counts, VA_DOUBLE * em, VA_DOUBLE * non_em, VA_INT32 * coverage, 
				VA_UINT32 ref_size , double gap_error, VA_INT32 * qval_array, VA_VecU8 * qval_ordering, int unique_qvals) {
#ifndef NDEBUG
	VA_ASSERT(seq.size==qvals.size);
#endif
	int j; int used_positions=0;
	for (j=0; j<seq.size; j++) {
		//If this is a funky letter or we have fallen off the END!
		if ( seq.data[j] == 5 || j + start > ref_size ) {
			continue;
		}
		coverage[ (start + j - 1) ] ++;
		//otherwise lets use this information in the covmap
		VA_DOUBLE error;
		if ( seq.data[j]==4) {
			error = gap_error;
		} else {
			error = qval_errors[qvals.data[j]];
			qval_array[(start+j-1)*(5*unique_qvals)+unique_qvals*seq.data[j]+qval_ordering->data[qvals.data[j]]]++;	
		}
		                   //row number     row size          column
		//update covmap counts
		//counts[(start+j-1)*5*2+5*reverse+seq.data[j]];
		//int matrix_index = (seq.data[j]) * ref_size + (start + j -1) ;
		int matrix_index = (start+j-1)*5 + seq.data[j];
#ifndef NDEBUG
		VA_ASSERT(matrix_index< ref_size*5);
#endif
		counts[matrix_index]++;
		//update covmap yes / non probabilities
#ifndef NDEBUG
		VA_ASSERT(error >= 0.0 && error <= 1.0);
#endif
		em[matrix_index] += log(1-error);
		non_em[matrix_index] += log(error/3);
		used_positions++;		
	}
	return used_positions;
}


Covmaps_t varid_covmaps(
    VA_MatI32 * letter_qvals,
    VA_MatI32 * color_qvals,
    VA_VecU8 * qval_ordering,
    int unique_qvals)
{
    Covmaps_t covmap;
    int usedreads = 0; /* the number of reads used */
    int i;
    size_t sz = reference_wgaps->size;
    size_t nrreads = realigned_reads.size;
    ReadsNode_t * rd = (ReadsNode_t *)realigned_reads.data;
    int zeroindexskippedreads = 0;


    /* coverage maps of various kind */
    covmap.coverage = va_allocMatI32(sz, 2, "covmap.coverage", __FILE__ ,__LINE__ );  
    memset(covmap.coverage.data, 0, sizeof(VA_INT32) * sz * 2);

    /* color emissions, in log space */
    /* histogram of colors */
    covmap.colors_positive = va_allocMatDouble(sz, 5,"covmap.colors_positive",__FILE__,__LINE__);  
    memset(covmap.colors_positive.data, 0, sizeof(VA_DOUBLE) * sz * 5);
    covmap.colors_negative = va_allocMatDouble(sz, 5,"covmap.colors_negative",__FILE__,__LINE__);  
    memset(covmap.colors_negative.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    /* col_emission */
    covmap.col_em = va_allocMatDouble(sz, 5,"covmap.col_em",__FILE__,__LINE__);  
    memset(covmap.col_em.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    /* noncol_emission */
    covmap.noncol_em = va_allocMatDouble(sz, 5,"covmap.noncol_em",__FILE__,__LINE__);  
    memset(covmap.noncol_em.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    /* letter emissions, in log space */
    /* histogram of letters */
    covmap.letters_positive = va_allocMatDouble(sz, 5,"covmap.letters_positive",__FILE__,__LINE__);  
    memset(covmap.letters_positive.data, 0, sizeof(VA_DOUBLE) * sz * 5);
    covmap.letters_negative = va_allocMatDouble(sz, 5,"covmap.letters_negative",__FILE__,__LINE__);  
    memset(covmap.letters_negative.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    /* let_emission */
    covmap.let_em = va_allocMatDouble(sz, 5,"covmap.let_em",__FILE__,__LINE__);  
    memset(covmap.let_em.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    /* nonlet_emission */
    covmap.nonlet_em = va_allocMatDouble(sz, 5,"covmap.nonlet_em",__FILE__,__LINE__);  
    memset(covmap.nonlet_em.data, 0, sizeof(VA_DOUBLE) * sz * 5);

    for (i = 0; i < nrreads; i++) {
        VA_DOUBLE * error_mat;
	if ( rd[i].isLetterspace ) {
		error_mat = error_mat_ls;
	} else {
		error_mat = error_mat_cs;
	}
		
	int used_positions = 0;
	
	/*if ( (rd[i].ls_seq.size!=1 || rd[i].cs_seq.size=0) && (rd[i].ls_seq.size==0 || rd[i].cs_seq.size!=0) ) {
		fprintf(stderr,"Read is neither color space or letter space? Offending read name: %s\n",rd[i].name);
		fprintf(stderr,"ls_seq_size %d, cs_seq_size %d\n",rd[i].ls_seq.size,rd[i].cs_seq.size);
		exit(1);
	}*/
#ifndef NDEBUG
	VA_ASSERT(rd[i].ls_seq.size == rd[i].ls_qvals.size);
#endif
	used_positions += update_map_with_read(rd[i].ls_seq, rd[i].ls_qvals,
				rd[i].ls_chrstart, error_mat,
				rd[i].strand=='-' ? covmap.letters_negative.data : covmap.letters_positive.data, covmap.let_em.data, covmap.nonlet_em.data, covmap.coverage.data, 
				reference_wgaps->size, gap_error, letter_qvals->data, qval_ordering, unique_qvals);
#ifndef NDEBUG
	VA_ASSERT(rd[i].cs_seq.size == rd[i].cs_qvals.size);
#endif
	used_positions += update_map_with_read(rd[i].cs_seq, rd[i].cs_qvals,
				rd[i].cs_chrstart, error_mat,
				rd[i].strand=='-' ? covmap.colors_negative.data : covmap.colors_positive.data, covmap.col_em.data, covmap.noncol_em.data, covmap.coverage.data,
				reference_wgaps->size, gap_error, color_qvals->data, qval_ordering, unique_qvals);

        if (used_positions == 0) {
            zeroindexskippedreads++;
        } else {
            usedreads++;
	}
    }

    fprintf(stderr, "   * %d/%lu reads used in coverage map, Skipped %d\n",usedreads, realigned_reads.size, zeroindexskippedreads);

    /*FILE* letters_file = fopen("letters.txt","w");
    //print_genome(letters_file,cdata.ref);
    print_matrix_double_raw(letters_file,covmap.letters.data, sz, 5);*/
    //FILE* colors_file = fopen("colors.txt","w");
    //print_genome(colors_file,cdata.ref);
    //print_matrix_double_raw(colors_file,covmap.colors_positive.data, sz, 5);
    /*FILE* colem_file = fopen("colem.txt","w");
    //print_genome(colem_file,cdata.ref);
    print_matrix_double_raw(colem_file,covmap.col_em.data, sz, 5);
    FILE* letem_file = fopen("letem.txt","w");
    //print_genome(letem_file,cdata.ref);
    print_matrix_double_raw(letem_file,covmap.let_em.data, sz, 5);
    FILE* non_colem_file = fopen("noncol_em.txt","w");
    //print_genome(non_colem_file,cdata.ref);
    print_matrix_double_raw(non_colem_file,covmap.noncol_em.data, sz, 5);
    FILE* non_letem_file = fopen("nonlet_em.txt","w");
    //print_genome(non_letem_file,cdata.ref);
    print_matrix_double_raw(non_letem_file,covmap.nonlet_em.data, sz, 5);
    FILE* coverage_file = fopen("coverage.txt","w");
    //print_genome(coverage_file,cdata.ref);
    print_matrix_int_raw(coverage_file,covmap.coverage.data, sz, 2);*/
    

    return (covmap);
} 


