/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"
#include <assert.h>





static double logfact_double(
	double x);

static int state_info(int state, VA_UINT8 * letter, VA_UINT8 * color, VA_UINT8 *, char * text);

static double poisson_penalty(const double * data_positive, const double * data_negative, const int hap1, const int hap2) {
	double c1 = data_positive[hap1]+data_negative[hap1];
	double c2 = data_positive[hap2]+data_negative[hap2];
	double sum = c1+c2;
	//double sum = data[0]+data[1]+data[2]+data[3]+data[4];
	double coverage = sum/2;
	double penalty=0.0;
	double c1_penalty = (c1==0.0 && coverage==0.0) ? 0.0 : c1*log(coverage);
	penalty+= c1_penalty - coverage - logfact_double(c1);
	double c2_penalty = (c2==0.0 && coverage==0.0) ? 0.0 : c2*log(coverage);
	penalty+= c2_penalty - coverage - logfact_double(c2);
	return penalty;
}

static double rank_sum_penalty(const int32_t * minor_allele,const int32_t * major_allele, const int unique_qvals) {
	int i; double n=0.0; double m=0.0;
	for (i=0; i<unique_qvals; i++) {
		n+=minor_allele[i];
		m+=major_allele[i];
	}
	double tie_break_factor=(n+m)*((n+m)*(n+m)-1)/12;
	double u=n*m+n*(n+1)/2; //This is how ALGLIB does it?
	double rank=1;
	for (i=0; i<unique_qvals; i++) {
		double range=minor_allele[i]+major_allele[i];
		u-=( (range+1)/2 +(rank-1))*((double)minor_allele[i]);
		tie_break_factor-=range*(range*range-1)/12;
		rank+=range;
	}
	double mu = n*m/2;
	//fprintf(stderr,"m %e, n %e, tie_b_f %e\n",m,n,tie_break_factor);
	double sigma = sqrt(m*n*tie_break_factor/((n+m)*(n+m-1)));
	if (n<2 || m<2) {
		return 0.0;
	}
	if ( (u-mu)==0 ) {
		return log(0.5);
	}
	double s = (u-mu)/sigma;
	//fprintf(stderr,"sigma %e, s %e, u %e, mu %e, %e vs %e\n",sigma,s,u,mu,(s>0 ? -s : s)/sqrt(2),0.5*(1+erf( (s>0 ? -s : s)/sqrt(2))));
	return log(0.5*(1+erf( (s>0 ? -s : s)/sqrt(2))));
}

static double binomial_penalty(const double * data_positive, const double * data_negative, const int hap1, const int hap2) {
	double c1 = data_positive[hap1]+data_negative[hap1];
	double c2 = data_positive[hap2]+data_negative[hap2];
	return logfact_double(c1+c2) - logfact_double(c1)- logfact_double(c2) - (c1+c2)*log(2);
}


static double normal_penalty_cov(const double * data_positive,const  double * data_negative,const  int hap1, const int hap2,const double cov) {
	double c1 = data_positive[hap1]+data_negative[hap1];
	double c2 = data_positive[hap2]+data_negative[hap2];
	double mean = (cov+1.0)/2.0;
	double penalty=0.0;
	//double std = (MIN(c1,c2)+1)/2.0;
	//double std = sqrt(0.0625*mean/2);
	//double std = sqrt(0.0625*mean);
	//double std = sqrt(0.0625*mean*0.625);
	double std = ((normal_cov_scale-1.0)/normal_cov_scale)*sqrt(normal_cov_scale*mean/(8.0*4.5));
	//double std = sqrt(0.1250*mean*2);
	penalty += -(c1-mean)*(c1-mean)/(2*std*std)-0.5*(log(M_PI)+2*log(std));
	penalty += -(c2-mean)*(c2-mean)/(2*std*std)-0.5*(log(M_PI)+2*log(std));
#ifndef NDEBUG
	if (isnan(penalty)) {
		fprintf(stderr,"normal_penalty cov : got a nan!\n");
		exit(1);
	}
#endif
	return MIN(0.0,penalty);
}

static double normal_penalty(const double * data_positive, const double * data_negative,const  int hap1, const int hap2) {
	double c1 = data_positive[hap1] + data_negative[hap1];
	double c2 = data_positive[hap2] + data_negative[hap2];
	double mean = (c1+c2)/2;
	double penalty=0.0;
	penalty += -(c1-mean)*(c1-mean)/(2*normal_std*normal_std)-0.5*(log(M_PI)+2*log(normal_std));
	penalty += -(c2-mean)*(c2-mean)/(2*normal_std*normal_std)-0.5*(log(M_PI)+2*log(normal_std));
#ifndef NDEBUG
	if (isnan(penalty)) {
		fprintf(stderr,"normal_penalty : got a nan!\n");
		exit(1);
	}
#endif
	return penalty;
}
/* map for the fasterfactorial function */
double factmap[] = {
    0,                 0,                 0.006931471805599, 0.017917594692281, 
    0.031780538303479, 0.047874917427820, 0.065792512120101, 0.085251613610654, 
    0.106046029027452, 0.128018274800815, 0.151044125730755, 0.175023078458739,
    0.199872144956619, 0.225521638531234, 0.251912211827387, 0.278992713838409, 
    0.306718601060807, 0.335050734501369, 0.363954452080330, 0.393398841871995, 
    0.423356164607535, 0.453801388984769, 0.484711813518352, 0.516066755677644,
    0.547847293981123, 0.580036052229805, 0.612617017610020, 0.645575386270063, 
    0.678897431371815, 0.712570389671680, 0.746582363488302, 0.780922235533153, 
    0.815579594561150, 0.850544670175815, 0.885808275421977, 0.921361756036871,
    0.957196945421432, 0.993306124547874, 1.029681986145138, 1.066317602606435, 
    1.103206397147574, 1.140342117814617, 1.177718813997451, 1.215330815154386, 
    1.253172711493569, 1.291239336391272, 1.329525750356163, 1.368027226373264,
    1.406739236482343, 1.445657439463449, 1.484777669517730, 1.524095925844974, 
    1.563608363030788, 1.603311282166309, 1.643201122631952, 1.683274454484277, 
    1.723527971391628, 1.763958484069973, 1.804562914175438, 1.845338288614495,
    1.886281734236716, 1.927390472878449, 1.968661816728900, 2.010093163992815, 
    2.051681994826412, 2.093425867525368, 2.135322414945633, 2.177369341139543, 
    2.219564418191304, 2.261905483237277, 2.304390435657770, 2.347017234428183,
    2.389783895618343, 2.432688490029827, 2.475729140961869, 2.518904022097232, 
    2.562211355500095, 2.605649409718632, 2.649216497985528, 2.692910976510198, 
    2.736731242856938, 2.780675734403661, 2.824742926876305, 2.868931332954269,
    2.913239500942703, 2.957666013507606, 3.002209486470141, 3.046868567656687, 
    3.091641935801469, 3.136528299498790, 3.181526396202093, 3.226634991267262, 
    3.271852877037753, 3.317178871969285, 3.362611819791985, 3.408150588707990,
    3.453794070622669, 3.499541180407702, 3.545390855194408, 3.591342053695754
};

/******************************************************************************
*   Here we compute the emission probabilities for each state along the
*       genome. We do this by considering the covmaps: For example, for
*       a state that supports color 3, we consider covmap.col_em for 3
*       and covmap.noncol_em for 0,1,2,4(gap). More details and parameter
*       information to come
*
* TODO: carefully reevaluate what penalties we're applying here...
*******************************************************************************/


VA_MatDouble varid_emissions(
    Covmaps_t    covmap,
    VA_MatI32 * letter_qvals,
    VA_MatI32 * color_qvals,
    VA_VecU8 * qval_ordering,
    int unique_qvals)
{
    int genome_length = reference_wgaps->size;

    fprintf(stderr,"   * SNP penalty %e, log of this is %e\n",snp_rate,log(snp_rate));
    //go through the whole genome
    VA_MatDouble e;
    e = va_allocMatDouble(genome_length, 1600,"e",__FILE__,__LINE__);
    memset(e.data,0,sizeof(VA_DOUBLE)*e.size_x * e.size_y);
    //go through the whole genome
    #pragma omp parallel
    {
    int i;
    #pragma omp for
    for (i=0; i<genome_length; i++) {
	int hap1_index;	
	for (hap1_index=0; hap1_index<40; hap1_index++)  {
		VA_UINT8 hap1_letter,hap1_previous,hap1_color; char hap1_text[3];
		state_info(hap1_index,&hap1_letter,&hap1_color,&hap1_previous,hap1_text);

		int hap2_index;
		for (hap2_index=0; hap2_index<40; hap2_index++) {
			VA_UINT8 hap2_letter,hap2_color,hap2_previous; char hap2_text[3];
			state_info(hap2_index,&hap2_letter,&hap2_color,&hap2_previous,hap2_text);
			int state=hap1_index*40 + hap2_index;
			int e_index = i*1600 + state;
			//PUT IN FREQ penalties
			if (hap1_letter!=4) {
				e.data[e_index]+=log(freq[hap1_letter]);
			}
			if (hap2_letter!=4) {
				e.data[e_index]+=log(freq[hap2_letter]);
			}

			//PUT IN SNP penalties
			int coverage[5] = {0,0,0,0,0};
			coverage[hap1_letter]++;
			coverage[hap2_letter]++;
			coverage[reference_wgaps->number_sequence[i]]++;
			int alleles = (coverage[0] > 0 ? 1 : 0) + (coverage[1] > 0 ? 1 : 0) + (coverage[2] > 0 ? 1 : 0) + (coverage[3] > 0 ? 1 : 0);
			if (alleles == 3) {
				//TRI ALLELIC SNP
				// (P+2P)^2 = 2*log(3P)
				e.data[e_index]+=2*log(3*snp_rate);
			} else if (alleles == 2) {
				if (hap1_letter != hap2_letter) {
					//DI ALLELIC SNP - heterozygous donor
					// 2P = log(2P)
					e.data[e_index]+=log(2*snp_rate);
				} else {
					//DI ALLELIC SNP - homozygous
					e.data[e_index]+=log(snp_rate);
				}
			}

			//PUT IN DELETION PENALTIES
			if (i>0 && reference_wgaps->number_sequence[i]!=4 && reference_wgaps->number_sequence[i-1]==4) {
				if (hap1_letter==4 && hap2_letter==4) {
					/*
						R:  AA---CGT
					       H1:      ?-
					       H2:      ?-
					*/
					e.data[e_index]+=log(2*deletion_open);
				} else if (hap1_letter==4 || hap2_letter==4) {
					/*
						R:  AA---CGT
					       HA:      ?-
					       HB:      ?X
					*/
					e.data[e_index]+=log(deletion_open);
				} else {
					//no penalty
				}
			} else if (i>0 && reference_wgaps->number_sequence[i]!=4 && reference_wgaps->number_sequence[i-1]!=4) {
				if (hap1_letter==4 && hap2_letter==4) {
					/*
						R:  AACGTAAGGCCCC---GT
					       H1:      ?-
					       H2:      ?-
					*/
					if (hap1_previous!=4 && hap2_previous!=4) {
						e.data[e_index]+=log(2*deletion_open);
					} else if (hap1_previous!=4 || hap2_previous!=4) {
						e.data[e_index]+=log(deletion_open);
					} else {
						e.data[e_index]+=log(deletion_extend);
					}
				} else if (hap1_letter==4) { 	
					/*
						R:  AACGTAAGGCCCC---GT
					       H1:      ?-
					       H2:      ?X
					*/
					if (hap1_previous==4) {
						e.data[e_index]+=log(deletion_extend);
					} else {
						e.data[e_index]+=log(deletion_open);
					}	
				} else if (hap2_letter==4) { 	
					/*
						R:  AACGTAAGGCCCC---GT
					       H1:      ?X
					       H2:      ?-
					*/
					if (hap2_previous==4) {
						e.data[e_index]+=log(deletion_extend);
					} else {
						e.data[e_index]+=log(deletion_open);
					}	
				} else {
					//no penalty
				}
			}
			//PUT IN INSERTION PENALTIES
			if (i>0 && reference_wgaps->number_sequence[i]==4 && reference_wgaps->number_sequence[i-1]!=4) {
				if (hap1_letter!=4 && hap2_letter!=4) {
					/*
						R:  AAC---------GT
					       H1:    ?X
					       H2:    ?Y
					*/
					e.data[e_index]+=log(2*insert_open);
				} else if (hap1_letter!=4 || hap2_letter!=4) {
					/*
						R:  AAC---------GT
					       H1:    ?-
					       H2:    ?X
					*/
					e.data[e_index]+=log(insert_open);
				} else {
					//no penalty
				}
			} else if (i>0 && reference_wgaps->number_sequence[i]==4 && reference_wgaps->number_sequence[i-1]==4) {
				if (hap1_letter!=4 && hap2_letter!=4) {
					/*
						R:  AAC---------GT
					       H1:       ?X
					       H2:       ?Y
					*/
					if (hap1_previous==4 && hap2_previous==4) {
						e.data[e_index]+=log(2*insert_open);
					} else if (hap1_previous==4 || hap2_previous==4) {
						e.data[e_index]+=log(insert_open);
						e.data[e_index]+=log(insert_extend);
					} else {
						e.data[e_index]+=log(2*insert_extend);
					}
				} else if (hap1_letter!=4) { 	
					/*
						R:  AAC---------GT
					       H1:       ?-
					       H2:       ?X
					*/		
					if (hap1_previous!=4) {
						e.data[e_index]+=log(insert_extend);
					} else {
						e.data[e_index]+=log(insert_open);
					}	
				} else if (hap2_letter!=4) { 	
					/*
						R:  AAC---------GT
					       H1:       ?X
					       H2:       ?-
					*/		
					if (hap2_previous!=4) {
						e.data[e_index]+=log(insert_extend);
					} else {
						e.data[e_index]+=log(insert_open);
					}	
				}
			} else if (i==0 && reference_wgaps->number_sequence[i]==4) {
					//its ok?
			} else if (i==0 && reference_wgaps->number_sequence[i]!=4) {
				if (hap1_letter==4 && hap2_letter==4) {
					e.data[e_index]+=log(2*deletion_open);
				} else if (hap1_letter==4) {
					e.data[e_index]+=log(deletion_open);
				} else if (hap2_letter==4) {
					e.data[e_index]+=log(deletion_open);
				} else {
					//its ok?
				}
			}

			//Get in the letter space signal	
			int letter;
			for (letter=0; letter<=4; letter++ ){
				if (letter!=hap2_letter && letter!=hap1_letter) {
					e.data[e_index]+=log_coefficient_ls*covmap.nonlet_em.data[i*5+letter];
				} else {
					e.data[e_index]+=log_coefficient_ls*covmap.let_em.data[i*5+letter];
				}
			}
			//letter signal distribution penalty?
			if (hap1_letter!=hap2_letter) {
				if (poisson==1) {
					e.data[e_index] += log_coefficient_ls*poisson_penalty(covmap.letters_positive.data+i*5,covmap.letters_negative.data+i*5,hap1_letter,hap2_letter);
				} else if (normal==1) {
					e.data[e_index] += log_coefficient_ls*normal_penalty(covmap.letters_positive.data+i*5,covmap.letters_negative.data+i*5,hap1_letter,hap2_letter);
				} else if (rank_sum==1) {
					int minor_allele=hap1_letter;
					int major_allele=hap2_letter;
					if ( covmap.letters_positive.data[i*5+hap1_letter] + covmap.letters_negative.data[i*5+hap1_letter] >
					     covmap.letters_positive.data[i*5+hap2_letter] + covmap.letters_negative.data[i*5+hap2_letter]) {
						minor_allele=hap2_letter;
						major_allele=hap1_letter;
					}
					VA_INT32 * minor_allele_qvals = 
						letter_qvals->data+i*(5*unique_qvals)+unique_qvals*minor_allele;
					VA_INT32 * major_allele_qvals = 
						letter_qvals->data+i*(5*unique_qvals)+unique_qvals*major_allele;
					double letter_coverage =0; int j;
					for (j=0; j<5; j++) {
						letter_coverage+=covmap.letters_positive.data[i*5+j]+covmap.letters_negative.data[i*5+j];
					}
					double log_penalty = log_coefficient_ls*rank_sum_penalty(minor_allele_qvals,major_allele_qvals,unique_qvals);
#ifndef NDEBUG
					if (isnan(log_penalty)) {
						fprintf(stderr,"A fatal error has occured in emissions, nan!\n");
						exit(1);
					}
#endif
					e.data[e_index] += log_penalty;
					//e.data[e_index] += log_coefficient_ls*binomial_penalty(covmap.letters.data+i*5,hap1_letter,hap2_letter);
					//e.data[e_index] += log_coefficient_ls*poisson_penalty(covmap.letters.data+i*5,hap1_letter,hap2_letter);
					e.data[e_index] += log_coefficient_ls*normal_penalty_cov(covmap.letters_positive.data+i*5,covmap.letters_negative.data+i*5,hap1_letter,hap2_letter,letter_coverage);
				} else if (binomial==1) {
					e.data[e_index] += log_coefficient_ls*binomial_penalty(covmap.letters_positive.data+i*5,covmap.letters_negative.data+i*5,hap1_letter,hap2_letter);
				} else {
					fprintf(stderr,"No distribution penalty selected!\n");
					exit(1);
				}
			}
			//Get in the color space signal
			int color;
			for (color=0; color<=4; color++ ){
				if (color!=hap2_color && color!=hap1_color) {
					e.data[e_index]+=log_coefficient_cs*covmap.noncol_em.data[i*5+color];
				} else {
					e.data[e_index]+=log_coefficient_cs*covmap.col_em.data[i*5+color];
				}
			}
			//color signal distribution penalty
			if (hap1_color!=hap2_color) {
				if (poisson==1) {
					e.data[e_index] += log_coefficient_cs*poisson_penalty(covmap.colors_positive.data+i*5,covmap.colors_negative.data+i*5,hap1_color,hap2_color);
				} else if (normal==1) {
					e.data[e_index] += log_coefficient_cs*normal_penalty(covmap.colors_positive.data+i*5,covmap.colors_negative.data+i*5,hap1_color,hap2_color);
				} else if (rank_sum==1) {
					int minor_allele=hap1_color;
					int major_allele=hap2_color;
					if ( covmap.colors_positive.data[i*5+hap1_color] + covmap.colors_negative.data[i*5+hap1_color] >
					     covmap.colors_positive.data[i*5+hap2_color] + covmap.colors_negative.data[i*5+hap2_color]) {
						minor_allele=hap2_color;
						major_allele=hap1_color;
					}
					VA_INT32 * minor_allele_qvals = 
						color_qvals->data+i*(5*unique_qvals)+unique_qvals*minor_allele;
					VA_INT32 * major_allele_qvals = 
						color_qvals->data+i*(5*unique_qvals)+unique_qvals*major_allele;
					double color_coverage =0; int j;
					for (j=0; j<5; j++) {
						color_coverage+=covmap.colors_positive.data[i*5+j]+covmap.colors_negative.data[i*5+j];
					}
					double log_penalty = log_coefficient_ls*rank_sum_penalty(minor_allele_qvals,major_allele_qvals,unique_qvals);
#ifndef NDEBUG
					if (isnan(log_penalty)) {
						fprintf(stderr,"A fatal error has occured in emissions, nan!\n");
						exit(1);
					}
#endif
					e.data[e_index] += log_penalty;
					//e.data[e_index] += log_coefficient_cs*binomial_penalty(covmap.colors.data+i*5,hap1_color,hap2_color);
					//e.data[e_index] += log_coefficient_cs*poisson_penalty(covmap.colors.data+i*5,hap1_color,hap2_color);
					e.data[e_index] += log_coefficient_cs*normal_penalty_cov(covmap.colors_positive.data+i*5,covmap.colors_negative.data+i*5,hap1_color,hap2_color,color_coverage);
				} else if(binomial==1) {
					e.data[e_index] += log_coefficient_cs*binomial_penalty(covmap.colors_positive.data+i*5,covmap.colors_negative.data+i*5,hap1_color,hap2_color);
				} else {
					fprintf(stderr,"No distribution penalty selected!\n");
					exit(1);
				}
			}
		}
	}
    }
    }   
    


    return (e);
}    


static double logfact_double(double x){
	if (x<100) {
		return lgamma(x+1);
	}
	return x*log(x)-x+0.5*log(2 * M_PI * x);
}


static int state_info(int state, VA_UINT8 * letter, VA_UINT8 * color, VA_UINT8 * previous, char * text) {
	text[0]=state_list[state][0];
	text[1]=state_list[state][1];
	text[2]='\0';
	VA_VecU8 text_vector; text_vector.size=2; text_vector.data=(VA_UINT8*)text;
	VA_VecU8 numbers=nuc2nuc(text_vector);
	text[0]=tolower(text[0]); text[1]=tolower(text[1]);
	VA_VecU8 colors=nuc2col(text_vector);
	*letter=numbers.data[1];
	*previous=numbers.data[0];
	*color= (*letter == 4) ? 4 : colors.data[0]; 
	free(numbers.data);
	free(colors.data);
	text[0]=state_list[state][0];
	text[1]=state_list[state][1];
	text[2]='\0';
	return 0;
}
