/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "fasta_reader_lib.h"
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"
#include "config.h"



static int cmpSortI32(
   const void * rec1,
   const void * rec2);



/*******************************************************************************
* VARID_SINGLE prepare the data and run the VARiD algorithm.
*   varid_single(readsfile, reffile, don_name, thr_vec_p, path) will run
*   the varid algorithm after preparing the data. If the reference file
*   (mfa) will have more than one reference, the readsfile will be split
*   based on the references they are aligned to, in the folder <path>, and
*   the reads will be loaded piecewise based on the reference that is being
*   processed. The result will be outputed in prettybase format.
*
*   Author: Adrian V. Dalca
*   (adalca at mit dot edu)
*******************************************************************************/
void get_reference() {
    if (load_reference_filename==NULL) {
	//build the fasta
	fprintf(stderr,"   - Loading from fasta file, %s\n",reference_filename);
	reference=fasta_read_file(reference_file);
	if (reference==NULL) {
		fprintf(stderr,"Failed to load from fasta file!\n");
	}
	fclose(reference_file);
	if (contig_name!=NULL) {
		reference=fasta_pop_and_free(reference,contig_name);
		if (reference==NULL) {
			fprintf(stderr,"Failed to find contig %s in fasta file!\n",contig_name);
			exit(1);
		}
	}
	fasta_crop(reference,hmm_start,hmm_end);
       	fasta_make_number_sequence(reference);
    } else {
	fprintf(stderr,"   - Loading from binary file, %s\n",load_reference_filename);
    	reference = fasta_load_region(load_reference_file,contig_name,hmm_start,hmm_end);
	fasta_make_letter_sequence(reference);
	fclose(load_reference_file);
    }
    if (contig_name==NULL) {
	contig_name=reference->name;
    }
    hmm_start=reference->start;
    hmm_end=reference->end;
    if (predict_start==0) {
	predict_start=hmm_start;
    }
    if (predict_end==0) {
	predict_end=hmm_end;
    }
    if (reference->size!=reference->original_size) {
	fprintf(stderr,"   - Only considering subsequence, %s:%llu-%llu\n",contig_name,hmm_start,hmm_end);
    } else {
	fprintf(stderr,"   - Considering full sequence, %s:%llu-%llu\n",contig_name,hmm_start,hmm_end);	
    }

    return;
}

void get_reference_wgaps(VA_VecI32 global_insert_v) {
    reference_wgaps = (fasta*)malloc(sizeof(fasta));
    if (reference_wgaps==NULL) {
	fprintf(stderr,"error: varid_single, failed to allocate memory for reference_wgaps\n");
	perror("");
	exit(1);
    }
    memset(reference_wgaps,0,sizeof(fasta));
    fprintf(stderr,"Setting refernce_wgaps name %s\n",reference->name);
    reference_wgaps->name=strdup(reference->name);
    reference_wgaps->size=reference->size + global_insert_v.size;
    reference_wgaps->number_sequence = (unsigned char*)malloc(reference_wgaps->size);
    if (reference_wgaps->number_sequence==NULL) {
	fprintf(stderr,"error: varid_single, failed to allocate memory for number sequence for reference_wgaps\n");
	perror("");
	exit(1);
    }
    memset(reference_wgaps->number_sequence,0,reference_wgaps->size);
    int ref_index=0; int inserts_index=0;
    int i;
    for (i=0; i<reference_wgaps->size; i++) {
	if (inserts_index<global_insert_v.size && global_insert_v.data[inserts_index] == i-inserts_index) {
		reference_wgaps->number_sequence[i]=4;
		inserts_index++;
	} else {
		reference_wgaps->number_sequence[i]=reference->number_sequence[ref_index++];
	}
    }
    fasta_make_letter_sequence(reference_wgaps);
    fasta_make_colour_sequence(reference_wgaps);
    return;
}

void varid_single() {
    struct timespec time_start; struct timespec time_end; struct timespec time_diff;
    //give some info to user
    fprintf(stderr, " * Alignments filename:\t%s\n",alignments_filename);
    fprintf(stderr, " * Temporary filename :\t%s\n",remapped_filename);
  
    //Load the reference into memory, through whatever means required 
    fprintf(stderr, " - Loading reference ...\n");
    va_GetTime(&time_start);
    get_reference();
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    fprintf(stderr, " * HMM Range          :\t%s:%llu-%llu\n",contig_name,hmm_start>0 ? hmm_start : 1,hmm_end>0 ? hmm_end : reference->size);
    fprintf(stderr, " * Predict Range      :\t%s:%llu-%llu\n",contig_name,predict_start>0 ? predict_start : 1,predict_end>0 ? predict_end : reference->size);
    fprintf(stderr, " * Call threshold     :\t%e\n",call_threshold); 

    //Split the reads if needed
    fprintf(stderr, " - %s alignments ...\n",predict_start>0 ? "Shifting" : "Scanning");
    va_GetTime(&time_start);
    long long unsigned used_reads=split_reads();
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    //Loading reads back in 
    fprintf(stderr, " - %s alignments ...\n","Reading in");
    va_GetTime(&time_start);
    //Set up the qval, and global insert vectors
    VA_VecU8 qval_ordering = va_allocVecU8(255);
    memset(qval_ordering.data,0,sizeof(VA_UINT8)*255);
    VA_VecI32 global_insert_v; 
    //Read them back in
    get_reads(used_reads, &original_reads, &global_insert_v, &qval_ordering);
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
  
    //Sort the inserts into reference 
    fprintf(stderr, " - Computing reference and alignments with inserts ...\n");
    va_GetTime(&time_start);
    qsort(global_insert_v.data, global_insert_v.size, sizeof(VA_INT32), cmpSortI32);

    //extract needed reference and insert
    get_reference_wgaps(global_insert_v);

    //realign the reads
    realigned_reads = align_reads(original_reads, global_insert_v);
    //make sure not to get confused with old original reads vector
    original_reads.data = NULL;
    original_reads.size = 0;
    //free the global insert v if need be
    if (global_insert_v.size>0) {
    	free(global_insert_v.data);
	global_insert_v.data=NULL;
    }
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);


    //figure out the ordering of qvals
    int temp=0; int i;
    for (i=0; i<qval_ordering.size; i++) {
	temp+=qval_ordering.data[i];
    } 
    temp=0;
    for (i=0; i<qval_ordering.size; i++) {
	if (qval_ordering.data[i]==1) {
		qval_ordering.data[i]=temp++;
	}
    } 

    //run VARiD algo
    Maps_t    maps;
    VA_MatDouble p;
    SnpGroup_t snpgroup;
    fprintf(stderr, " - Running VARiD HMM prediction model ...\n");
    va_GetTime(&time_start);
    varid(0, &maps, &(snpgroup.covmap), &p, &qval_ordering);

    /*{
	FILE * f = fopen("p_matrix_before.txt","w");
	print_genome(f,g_ref);
	print_matrix_double(f,p.data,p.size_x,p.size_y);
	fclose(f);
    }*/
    predict(p, maps, &snpgroup.maxvals_2, 
	    &snpgroup.lmat, 
	    &snpgroup.cmat);
    free(maps.symm.data);
    free(maps.call_letter.data);
    free(maps.full_letter.data);
    free(maps.call_color.data);
    free(p.data);
    /*{
	FILE * f = fopen("p_matrix_after.txt","w");
	print_genome(f,g_ref);
	print_matrix_double(f,p.data,p.size_x,p.size_y);
	fclose(f);
    }*/

    /* prepare the SNP altering matrix */
    snpgroup.ch_list = va_allocMatU8(snpgroup.cmat.size_x, 1);
    memset(snpgroup.ch_list.data, 0,
           snpgroup.ch_list.size_x * snpgroup.ch_list.size_y * sizeof(VA_UINT8));

    /*destroy_state(cdata, snpgroup.data->cmat, g_reads, 
		  g_readstarts, snpgroup.data->covmap, thr_vec_snp, 
		  snpgroup.data->lmat, &snpgroup.data->ch_list, 
		  &tot, &snpgroup.data->top);*/
    /*int y;
    for (y=0; y< snpgroup.lmat.size_y; y++ ){
	int x;
	for (x=0; x< snpgroup.lmat.size_x; x++) {
		printf("%d",snpgroup.lmat.data[y*snpgroup.lmat.size_x + x]);
	}
	printf("\n");
	for (x=0; x< snpgroup.lmat.size_x; x++) {
		printf("%d",snpgroup.cmat.data[y*snpgroup.cmat.size_x + x]);
	}
	printf("\n");
    }*/


    //free the reads in memory
    ReadsNode_t * rd = (ReadsNode_t *)realigned_reads.data;
    int j;
    for (j = 0; j < realigned_reads.size; j++) {
	if (rd[j].name != NULL)
	    free(rd[j].name);
	if (rd[j].ls_seq.data != NULL)
	    free(rd[j].ls_seq.data);
	if (rd[j].ls_qvals.data != NULL)
	    free(rd[j].ls_qvals.data);
	if (rd[j].cs_seq.data != NULL)
	    free(rd[j].cs_seq.data);
	if (rd[j].cs_qvals.data != NULL)
	    free(rd[j].cs_qvals.data);
	if (rd[j].ins.data != NULL)
	    free(rd[j].ins.data);
	if (rd[j].cs_ins.data != NULL)
	    free(rd[j].cs_ins.data);
	if (rd[j].ls_ins.data != NULL)
	    free(rd[j].ls_ins.data);
    }
    free(rd);

	VA_VecU8 pred1_whole;
	VA_VecU8 pred2_whole;

	get_pred(snpgroup.covmap, maps,
		 snpgroup.ch_list,
		 snpgroup.maxvals_2, snpgroup.lmat,
		 &pred1_whole, &pred2_whole);
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    /*{
    int y;
    for (y=0; y< snpgroup.data[gnr].ch_list.size_y; y++ ){
	int x;
	for (x=0; x< snpgroup.data[gnr].ch_list.size_x; x++) {
		printf("%d",snpgroup.data[gnr].ch_list.data[y*snpgroup.data[gnr].ch_list.size_x + x]);
	}
	printf("\n");
    }
    printf("Pred1\n");
    for (y=0; y<pred1_whole.size; y++) {
	printf("%d",pred1_whole.data[y]);
    }
    printf("\nPred2\n");
    for (y=0; y<pred2_whole.size; y++) {
	printf("%d",pred2_whole.data[y]);
    }
    printf("\n");
    }*/
	if (format==FORMAT_PB) {
		fprintf(stderr," - Writting pretty base output ...\n");
		va_GetTime(&time_start);
		writePrettybase(pred1_whole, pred2_whole, snpgroup.maxvals_2);
	} else if (format==FORMAT_VCF) {
		fprintf(stderr," - Writting VCF output ...\n");
		va_GetTime(&time_start);
		writeVCF(pred1_whole, pred2_whole, snpgroup.maxvals_2,
			    snpgroup.covmap,snpgroup.cmat,
			    snpgroup.lmat);

	} else {
		fprintf(stderr,"An invalid output format has been selected!\n");
		exit(1);
	}

    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);
    fprintf(stderr," - Cleaning up a bit ...\n");
    va_GetTime(&time_start);
	free(pred1_whole.data);
	free(pred2_whole.data);

    free(snpgroup.lmat.data);
    free(snpgroup.cmat.data);
    free(snpgroup.ch_list.data);
    free(snpgroup.covmap.coverage.data);
    free(snpgroup.covmap.colors_positive.data);
    free(snpgroup.covmap.colors_negative.data);
    free(snpgroup.covmap.letters_positive.data);
    free(snpgroup.covmap.letters_negative.data);
    free(snpgroup.covmap.col_em.data);
    free(snpgroup.covmap.noncol_em.data);
    free(snpgroup.covmap.let_em.data);
    free(snpgroup.covmap.nonlet_em.data);
    free(snpgroup.maxvals_2.data);


    if (realignments_file!=NULL) {
	fclose(realignments_file);
    }
    if (read_alignments_file!=NULL ){
	fclose(read_alignments_file);
    }
    va_GetTime(&time_end);
    time_diff = va_TimeSub(&time_end, &time_start);
    fprintf(stderr, " + Done. %ld.%03ld secs\n",time_diff.tv_sec, time_diff.tv_nsec/NS_IN_MS);

}


static int cmpSortI32(
   const void * rec1,
   const void * rec2)
{
   VA_INT32 p1;
   VA_INT32 p2;

   VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

   p1 = *(VA_INT32 *)rec1;
   p2 = *(VA_INT32 *)rec2;

   return (p1 - p2);
}


