/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* compares two groups of snp results predictions
*   stats = compare_grp(rha_real, hap1, hap2, ref, covmap) compares the snp
*   predictions done via the hap1-hap2-ref results from a snp predictor
*   such as VARid, and compares it with rha_real. Several stats are 
*   provided in the output. Optinally, if covmap is provided, some detailed
*   stats can be provided. (TODO: not quite finished yet, depends on
*   debugging necessity)
*
*   Parameters:
*   rha_real - the last column of a prettybase file indicating the 'real'
*       calls at each position in a reference, with the letters 'R' (no
*       variation), 'H' (heterozygous variation), and 'A' (homozyous 
*       variation)
*   hap1 - the first haplome prediction, in ns letterspace (see NUC2NUC)
*   hap2 - the second haplome prediction, in ns letterspace 
*   ref - the reference, in ns letterspace
*   covmap_coverage - optional, the covmap.coverage field of covmap.
*    
*   Returns
*   stats - a struct with the following fields
*       fp - the false positive count
*       fpvec - the positions of the false positives
*       fn - the false negative count
*       fnvec - the positions of the false negatives
*       fo - other falses, hopefully these are variation calls that we
*           claim are A and are actually H, or v-v. 
*       fovec - the positions of such false calls
*       bad_calls - a boolean vector the size of the reference, on whether
*           the call is bad
*       covpos - the coverage at every bad call.
* 
*   See also PBRESSTATS, NUC2NUC
*
*   Author: Adrian V. Dalca
*******************************************************************************/
VA_InternalStats compare_grp(
    VA_Vector    rha_real, 
    VA_VecU8     hap1, 
    VA_VecU8     hap2, 
    VA_VecU8     ref, 
    Covmaps_t    covmap)
{
    int i;
    int j;
    VA_InternalStats stats;

    /* Prepare the data */
    VA_VecU8 don1 = nuc2nuc(hap1);
    VA_VecU8 don2 = nuc2nuc(hap2);
    VA_VecU8 ref2 = nuc2nuc(ref);
    VA_VecU8 don1_corr = va_allocVecU8(don1.size);
    VA_VecU8 don2_corr = va_allocVecU8(don2.size);
    VA_VecU8 ref_corr = va_allocVecU8(ref2.size);
    VA_VecU8 norefgap;
    int sz;
    VA_VecU8 rha_call_R1;
    VA_VecU8 rha_call_R2;
    VA_VecU8 rha_call_H;
    VA_VecU8 rha_call_A;
    VA_VecU8 rha_call;
    VA_VecU32 gap_loc;
    int gap_index;
    VA_VecU8 d1_tmp;
    VA_VecU8 d2_tmp;
    int nbad_calls = 0;
    char ** rha_r = (char **)rha_real.data;
    char * rha_real_tmp; 
    
    VA_ASSERT(don1.size = don2.size);
    VA_ASSERT(don1.size = ref2.size);

    don1_corr.size = 0;
    don2_corr.size = 0;
    ref_corr.size = 0;

    /* convert into prettybase formata */
    for (i = 0; i < don1.size; i++)
    {
        don1.data[i] = toupper(don1.data[i]);
        don2.data[i] = toupper(don2.data[i]);
        ref2.data[i] = toupper(ref2.data[i]);

        /* correct any positions with gaps in all three data */
        if ((ref2.data[i] != '-') ||
           (don1.data[i] != '-') ||
           (don2.data[i] != '-'))
        {
            don1_corr.data[don1_corr.size++] = don1.data[i];
            don2_corr.data[don2_corr.size++] = don2.data[i];
            ref_corr.data[ref_corr.size++] = ref2.data[i];
        }
    }

    if (don1_corr.size != don1.size)
    {
        free(don1.data);
        free(don2.data);
        free(ref2.data);

        memcpy(&don1, &don1_corr, sizeof(don1));
        memcpy(&don2, &don2_corr, sizeof(don2));
        memcpy(&ref2, &ref_corr, sizeof(ref2));
    }
    
    sz = 0;
    norefgap = va_allocVecU8(ref2.size);
    rha_call_R1 = va_allocVecU8(ref2.size);
    rha_call_R2 = va_allocVecU8(ref2.size);
    rha_call_H = va_allocVecU8(ref2.size);
    rha_call_A = va_allocVecU8(ref2.size);
    rha_call = va_allocVecU8(ref2.size);
    gap_loc = va_allocVecU32(ref2.size);
    gap_loc.size = 0;
   
    /* build the call rha_call. */
    for (i = 0; i < ref2.size; i++)
    {
        rha_call.data[i] = '-';

        if (ref2.data[i] != '-')
        {
            norefgap.data[i] = TRUE;
            sz++;
        }
        else
        {
            norefgap.data[i] = FALSE;
            gap_loc.data[gap_loc.size++] = i + 1;
        }

        rha_call_R1.data[i] = (norefgap.data[i] && 
                              (don1.data[i] == don2.data[i]) &&
                              (don1.data[i] == 'N')) ? TRUE : FALSE;
        rha_call_R2.data[i] = (norefgap.data[i] && 
                              (don1.data[i] == don2.data[i]) &&
                              (don1.data[i] == ref2.data[i])) ? TRUE : FALSE;
        rha_call_A.data[i] = (norefgap.data[i] && 
                              (don1.data[i] == don2.data[i]) &&
                              (don1.data[i] != 'N') &&
                              (don1.data[i] != ref2.data[i])) ? TRUE : FALSE;
        rha_call_H.data[i] = (norefgap.data[i] && 
                              (don1.data[i] != don2.data[i])) ? TRUE : FALSE;
        if (rha_call_R1.data[i] || rha_call_R2.data[i])
            rha_call.data[i] = 'R';
        if (rha_call_A.data[i])
            rha_call.data[i] = 'A';
        if (rha_call_H.data[i])
            rha_call.data[i] = 'H';
    }

    free(rha_call_R1.data);
    free(rha_call_R2.data);
    free(rha_call_A.data);
    free(rha_call_H.data);
    free(norefgap.data);

    /* verify size of data */
    VA_ASSERT(sz == rha_real.size);

    /* finish the rha_call vector for the (rare) gap situations */
    if (gap_loc.size == 0)
    {
        free(gap_loc.data);
        gap_loc.data = NULL;
    }
    else
    {
        gap_loc = va_reallocVecU32(gap_loc, gap_loc.size);
    }
    gap_index = 1;
    
    d1_tmp = va_allocVecU8(don1.size);
    d2_tmp = va_allocVecU8(don2.size);
 
    while (gap_index <= gap_loc.size)
    {
        int eql;
        int eql_N;
        int gap = gap_loc.data[gap_index - 1];
        
        d1_tmp.size = 0;
        d2_tmp.size = 0;
        d1_tmp.data[d1_tmp.size++] = don1.data[gap - 1];
        d2_tmp.data[d2_tmp.size++] = don2.data[gap - 1];
        i = 1;

        while ((gap_loc.size >= gap_index + i) && 
               (gap + i == gap_loc.data[gap_index + i - 1]))
        {
            d1_tmp.data[d1_tmp.size++] = don1.data[gap + i - 1];
            d2_tmp.data[d2_tmp.size++] = don2.data[gap + i - 1];
            i = i + 1;
        }
        
        eql = TRUE;
        eql_N = TRUE;
        for (j = 0; j < d1_tmp.size; j++)
        {
            if (d1_tmp.data[j] != d2_tmp.data[j])
                eql = FALSE;
            if ((d1_tmp.data[j] != '-') && (d1_tmp.data[j] != 'N')) 
                eql_N = FALSE;
        }

        if (eql && eql_N)
        {
           ;
        }
        else if (eql)
        {
            rha_call.data[gap - 1 - 1] = 'A';
        }
        else
        {
            rha_call.data[gap - 1 - 1] = 'H';
        }
        
        gap_index = gap_index + i;
    }
    
    j = 0;
    sz = 0;
    if (don1.size > 0)
    {
        for (i = 0; i < don1.size; i++)
        {
            if (j < gap_loc.size)
            {
                if (i + 1 == gap_loc.data[j])
                {
                    j++;
                    continue;
                }
            }

            don1.data[sz] = don1.data[i];
            don2.data[sz] = don2.data[i];
            rha_call.data[sz] = rha_call.data[i];
            VA_ASSERT(rha_call.data[sz] != '-');
            sz++;
        }

        don1 = va_reallocVecU8(don1, sz);
        don2 = va_reallocVecU8(don2, sz);
        rha_call = va_reallocVecU8(rha_call, sz);
    }
    
    /* 
     * Proceed with the comparison between rha_call (prediction) and
     * rha_real (real)
     */ 
    VA_ASSERT(rha_call.size == rha_real.size);
    stats.fp = va_allocVecU32(rha_call.size);
    stats.fp.size = 0;
    stats.fn = va_allocVecU32(rha_call.size);
    stats.fn.size = 0;
    stats.tp = va_allocVecU32(rha_call.size);
    stats.tp.size = 0;
    stats.tn = va_allocVecU32(rha_call.size);
    stats.tn.size = 0;
    stats.fo = va_allocVecU32(rha_call.size);
    stats.fo.size = 0;
    stats.bad_calls = va_allocVecU32(rha_call.size);

    sz = 0;
    for (i = 0; i < rha_real.size; i++)
        sz += strlen(rha_r[i]);

    rha_real_tmp = (char *)malloc(sz + 1);
    VA_ASSERT(rha_real_tmp != NULL);
    memset(rha_real_tmp, 0, sz + 1);

    sz = 0;
    for (i = 0; i < rha_real.size; i++)
    {
        size_t l = strlen(rha_r[i]);
        memcpy(rha_real_tmp + sz, rha_r[i], l);
        sz += l;
    }

    /* new vectorized method to get the stats */
    for (i = 0; i < rha_call.size; i++)
    {
        if (rha_call.data[i] != rha_real_tmp[i])
        {
            nbad_calls++;
            stats.bad_calls.data[i] = TRUE;
        }
        else
        {
            stats.bad_calls.data[i] = FALSE;
        }

        if ((rha_call.data[i] != 'R') && (rha_real_tmp[i] == 'R'))
            stats.fp.data[stats.fp.size++] = i + 1; 
        if ((rha_call.data[i] != 'R') && (rha_real_tmp[i] != 'R'))
            stats.tp.data[stats.tp.size++] = i + 1; 
        if ((rha_call.data[i] == 'R') && (rha_real_tmp[i] != 'R'))
            stats.fn.data[stats.fn.size++] = i + 1; 
        if ((rha_call.data[i] == 'R') && (rha_real_tmp[i] == 'R'))
            stats.tn.data[stats.tn.size++] = i + 1; 
        if ((rha_call.data[i] != 'R') && (rha_real_tmp[i] != 'R') &&
            stats.bad_calls.data[i])
            stats.fo.data[stats.fo.size++] = i + 1; 
    }
    
    VA_ASSERT(nbad_calls == stats.fp.size + stats.fn.size + stats.fo.size);
    
    free(rha_real_tmp);
    
    sz = 0;
    for (i = 0; i < stats.bad_calls.size; i++)
        if (stats.bad_calls.data[i])
            sz++;

    stats.covpos = va_allocMatU32(2, sz);
    stats.covpos.size_y = 0;
    for (i = 0; i < stats.bad_calls.size; i++)
    {
        if (stats.bad_calls.data[i])
        {
            size_t b = stats.covpos.size_x * stats.covpos.size_y;
            stats.covpos.data[b] = i + 1;
            stats.covpos.data[b + 1] = 
                covmap.coverage.data[i + 1];
            stats.covpos.size_y++;
        }
    }

    if (stats.fp.size == 0)
    {
        free(stats.fp.data);
        stats.fp.data = NULL;
    }

    if (stats.fn.size == 0)
    {
        free(stats.fn.data);
        stats.fn.data = NULL;
    }

    if (stats.fo.size == 0)
    {
        free(stats.fo.data);
        stats.fo.data = NULL;
    }

    if (stats.covpos.size_y == 0)
    {
        free(stats.covpos.data);
        stats.covpos.data = NULL;
        stats.covpos.size_x = 0;
    }
        
    return (stats);
}

