/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
//#include <malloc.h>
#include "va_assert.h"
#include "va_log.h"
#include "varid.h"


static int compareInsvec(
    const void * rec1,
    const void * rec2);


/******************************************************************************
*   extract the reads and populates the necessary struct vector
*   [reads, insvec] = GET_READS(mapfile) will read the mapping in file
*       mapfile and populate the struct vector reads, as well as the
*       insertion vector insvec.
*
*   [reads, insvec] = GET_READS(mapfile, nr_reads, nr_lines) will proceed
*   similarly, however by supplying the number of reads and the number of
*   lines we can achieve better storage and get a waitbar.
*
*   Parameters:
*   mapfile - path to the mapped reads file (probcalc format)
*
*   Optional:
*   nr_reads - if given, will be able to estimate struct sizes correctly.
*   nr_lines - if given, can run waitbar
*
*   Returns:
*   reads  - a (nr_reads) vector of structs with the fields indicated in
*       varid_const.m
*
*   Author: Adrian V. Dalca
*******************************************************************************/
void get_reads(
    long long unsigned used_reads, 
    VA_Vector  * outreads,
    VA_VecI32  * outinsvec,
    VA_VecU8 * qval_ordering) {

    VA_ASSERT(outreads != NULL);
    VA_ASSERT(outinsvec != NULL);

    ReadsNode_t * reads = (ReadsNode_t *)malloc(sizeof(ReadsNode_t) * used_reads);
    VA_ASSERT(reads != NULL);
    memset(reads, 0, sizeof(ReadsNode_t) * used_reads);

    VA_VecI32 insvec;
    memset(&insvec, 0, sizeof(insvec));

    FILE * remapped_file = fopen(remapped_filename, "r");
    if (remapped_file==NULL) {
	fprintf(stderr,"error: get_reads, failed to open remapped file, %s for reading\n",remapped_filename);
	perror("");
	exit(1);
    }
    char sam_line[BUFSIZ];
    int i;
    for(i=0; fgets(sam_line,BUFSIZ,remapped_file) && i<used_reads; i++) {
	if (sam_line[0]!='@') {
        	process_read(sam_line, reads+i, &insvec,qval_ordering);
	} else {
		i--;
	}
	
    }
    VA_ASSERT(i==used_reads);

    outreads->data = reads;
    outreads->size = used_reads;
    if (insvec.size > 0) {
        qsort(insvec.data, insvec.size, sizeof(VA_INT32), compareInsvec); 
        outinsvec->data = insvec.data;
        outinsvec->size = insvec.size;
    } else {
	outinsvec->data = NULL;
	outinsvec->size = 0;
    }

    fclose(remapped_file);

    learn_error();
}

static int compareInsvec(
    const void * rec1,
    const void * rec2) {
    VA_INT32 p1;
    VA_INT32 p2;
#ifndef NDEBUG
    VA_ASSERT((rec1 != NULL) && (rec2 != NULL));
#endif
    p1 = *(VA_INT32 *)rec1;
    p2 = *(VA_INT32 *)rec2;

    return (p1 - p2);
}
        

