#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>
#include <arpa/inet.h>
#include "fasta_reader_lib.h"

#define FASTA_READER_BUFFER_SIZE	1024*30

static char buffer[FASTA_READER_BUFFER_SIZE];
static char next_line[FASTA_READER_BUFFER_SIZE];
static size_t buffer_full=0;
static size_t buffer_index=0;

static int unused_line=0;

static char * get_next_line(FILE * f) {
	if (unused_line) {
		unused_line=0;
		return next_line;
	}
	//Check if we need to read in more
	if (buffer_full==buffer_index) {
		buffer_full=fread(buffer,1,FASTA_READER_BUFFER_SIZE,f);
		buffer_index=0;
		if (buffer_full==0) {
			return NULL;
		}
	}
	//Find the end of the first line
	int i;
	for (i=0; (i+buffer_index)<buffer_full && buffer[i+buffer_index]!='\n'; i++);
	if ((i+buffer_index)==buffer_full) {
		if (buffer_index!=0) {
			int j;
			for (j=0; j<(buffer_full-buffer_index); j++ ){
				buffer[j]=buffer[j+buffer_index];
			}
			buffer_full-=buffer_index;
			buffer_index=0;
		}
		size_t temp = fread(buffer+buffer_full,1,FASTA_READER_BUFFER_SIZE-buffer_full,f);
		buffer_full+=temp;
		for (i=0; (i+buffer_index)<buffer_full && buffer[i+buffer_index]!='\n'; i++);
	}	
	if ((i+buffer_index)==buffer_full) {
		fprintf(stderr,"Maximum line length exceeded!\n");
		exit(1);
	}
	//Copy over the line
	assert((buffer_index+i)<FASTA_READER_BUFFER_SIZE);
	int j;
	for (j=0; j<i; j++) {
		next_line[j]=buffer[buffer_index+j];
	}
	next_line[j]='\0';
	buffer_index+=j+1;
	return next_line;
}

fasta * fasta_read_file(FILE * f) {
	fasta * head; fasta * current; fasta * previous;
	previous=head=fasta_read_next(f);
	while((current=fasta_read_next(f))) {
		previous->next=(struct fasta *)current;
		previous=current;
	}
	return head;
}

fasta * fasta_read_next(FILE * f) {
	char * line = get_next_line(f);
	while (line!=NULL && line[0]=='\0') {
		line=get_next_line(f);
	}
	if (line==NULL) {
		return NULL;
	}
	//Grab the header
	assert(line[0]=='>');
	line++;
	int i;
	for (i=0; line[i]!='\t' && line[i]!=' ' && line[i]!='\0'; i++);
	char * header = (char*)malloc(i+1);
	if (header==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for header\n");
		exit(1);
	}
	for (i=0; line[i]!='\t' && line[i]!=' ' && line[i]!='\0'; i++) {
		header[i]=line[i];
	}
	header[i]='\0';
	//Get the sequence
	char * sequence = (char *)malloc(FASTA_READER_BUFFER_SIZE);
	if (sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for sequence\n");
		exit(1);
	}
	size_t sequence_size = 0; size_t sequence_memory = FASTA_READER_BUFFER_SIZE;
	line=get_next_line(f);
	while (line!=NULL && line[0]!='>') {
		size_t size = strlen(line);
		while (sequence_size+size+1>=sequence_memory) {
			sequence_memory=2*sequence_memory;
			sequence=realloc(sequence,sequence_memory);
			if (sequence==NULL) {
				fprintf(stderr,"error: fasta_reader_lib failed to reallocate memory for sequence\n");	
				exit(1);
			}
		}
		size_t i;
		for (i=0; i<size; i++) {
			sequence[sequence_size++]=line[i];
		}
		line=get_next_line(f);	
	}
	sequence[sequence_size]='\0';
	if (line!=NULL && line[0]=='>') {
		unused_line=1;
	}
	sequence=realloc(sequence,sequence_size+1);
	if (sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to reallocate memory for sequence, 2");
		exit(1);
	}
	fasta * fa = (fasta*)malloc(sizeof(fasta));
	if (fa==NULL ){
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for fasta structure");
		exit(1);
	}
	memset(fa,0,sizeof(fasta));
	fa->sequence=sequence;
	fa->size=sequence_size;
	fa->original_size=sequence_size;
	fa->start=1;
	fa->end=fa->size;
	fa->name=header;
	return fa;
}


void fasta_make_number_sequence(fasta * fa) {
	if (fa->sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib sequence must be present to make number sequence\n");
		exit(1);
	}
	if (fa->size==0) {
		fprintf(stderr,"error: fasta_reader_lib make_number_sequence, sequence must have positive length\n");
		exit(1);
	}
	unsigned char * number_sequence = (unsigned char *)malloc(sizeof(unsigned char)*fa->size);
	if (number_sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for number sequence\n");
		exit(1);
	}
	int i;
	for (i=0; i<fa->size; i++) {
		switch(fa->sequence[i]) {
			case 'A':
			case 'a':
				number_sequence[i]=0;
				break;
			case 'C':
			case 'c':
				number_sequence[i]=1;
				break;
			case 'G':
			case 'g':
				number_sequence[i]=2;
				break;
			case 'T':
			case 't':
				number_sequence[i]=3;
				break;
			case '-':
			case ' ':
				number_sequence[i]=4;
				break;
			case 'N':
			case 'n':
				number_sequence[i]=5;
				break;
			default:
				fprintf(stderr,"Failed to parse character %c in make_number_sequence, assuming 'N'\n",fa->sequence[i]);
				number_sequence[i]=5;
				break;
		}
	}
	fa->number_sequence=number_sequence;
}

void fasta_make_letter_sequence(fasta * fa) {
	if (fa->number_sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib number sequence must be present to make letter sequence\n");
		exit(1);
	}
	if (fa->size==0) {
		fprintf(stderr,"error: fasta_reader_lib make_letter_sequence, sequence must have positive length\n");
		exit(1);
	}
	char * letter_sequence = (char *)malloc(fa->size+1);
	if (letter_sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for letter sequence\n");
		exit(1);
	}
	int i;
	for (i=0; i<fa->size; i++) {
		switch(fa->number_sequence[i]) {
			case 0:
				letter_sequence[i]='a';
				break;	
			case 1:
				letter_sequence[i]='c';
				break;
			case 2:
				letter_sequence[i]='g';
				break;
			case 3:
				letter_sequence[i]='t';
				break;
			case 4:
				letter_sequence[i]='-';
				break;
			case 5:
				letter_sequence[i]='n';
				break;
			default:	
				fprintf(stderr,"Failed to parse number %u in make_letter_sequence, assuming '5'\n",fa->number_sequence[i]);
				letter_sequence[i]='n';
				break;
		}
	}
	letter_sequence[i]='\0';
	fa->sequence=letter_sequence;
}


void fasta_make_colour_sequence(fasta * fa) {
	unsigned char * colour_sequence = (unsigned char *)malloc(sizeof(unsigned char)*fa->size);
	if (colour_sequence==NULL) {
		fprintf(stderr,"error: fasta_reader_lib failed to allocate memory for number sequence\n");
		exit(1);
	}
	colour_sequence[0]=6;
	unsigned char last=fa->number_sequence[0];
	int i;
	for (i=1; i<fa->size; i++) {
		unsigned char current=fa->number_sequence[i];
		assert(current<6);
		switch(current) {
			case 4:
				colour_sequence[i]=4;
				break;
			case 5:
				colour_sequence[i]=5;
				last=5;
				break;
			default:
				if (last==5) {
					colour_sequence[i]=5;
				} else {
					colour_sequence[i]=last^current;
				}
				last=current;
				break;	
		}
	}
	fa->colour_sequence=colour_sequence;	
}


void fasta_save(fasta * head, FILE * f) {
	fasta * current;
	current=head;
	size_t length=0;
	while(current!=NULL) {
		length++;
		current=(fasta *)current->next;
	}
	size_t lengths[length];
	current=head;
	size_t i;
	size_t length_so_far=4+4*length;
	for (i=0; i<length; i++) {
		lengths[i]=length_so_far;
		length_so_far+=current->size+strlen(current->name)+1;
		current=(fasta *)current->next;
	}
	uint64_t max_32bit=4294967295;
	if (length>max_32bit) {
		fprintf(stderr,"error: fasta_reader fasta_save, number of fasta entries must be less then %lu\n",max_32bit);
		exit(1);
	}
	if (length_so_far>max_32bit) {
		fprintf(stderr,"error: fasta_reader fasta_save, cumulative length of sequences, titles and overhead must be less then %lu\n",max_32bit);
		exit(1); 
	}
	uint32_t length32 = htonl((uint32_t) length);
	uint32_t lengths32[length];
	for (i=0; i<length; i++) {
		lengths32[i]=htonl((uint32_t)lengths[i]);
	}
	assert(sizeof(uint32_t)==4);
	//write the header of the file
	size_t wrote = fwrite(&length32, 4, 1, f);
	wrote+=fwrite(lengths32,4,length,f);
	if (wrote!=length+1) {
		fprintf(stderr,"error: fasta_reader fasta_save, failed to write output\n");
		perror("");
		exit(1);
	}
	//write the data
	current=head;
	while (current!=NULL) {
		if (current->start!=1 || current->original_size!=current->end) {
			fprintf(stderr,"error: fasta_reader fasta_save, must write full sequence to binary! not a subset!\n");
			exit(1);
		}
		//write the length of header
		uint32_t name_length = strlen(current->name)+1;
		uint32_t name_length32 = htonl((uint32_t)name_length); 
		wrote=fwrite(&name_length32, 4, 1, f);
		wrote+=fwrite(current->name,1,name_length,f);
		if (wrote!=name_length+1) {
			fprintf(stderr,"error: fasta_reader fasta_save, failed to write header in output, %lu vs %u\n",wrote,name_length+1);
			perror("");
			exit(1);
		}
		//write the number sequence out
		if (current->number_sequence==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_save, failed to write number sequence because it is NULL!\n");
			exit(1);
		}
		uint32_t sequence_length32 = htonl((uint32_t)current->size);
		wrote=fwrite(&sequence_length32,4,1,f);
		wrote+=fwrite(current->number_sequence,current->size,sizeof(unsigned char),f);
		if (wrote!=2) {
			fprintf(stderr,"error: fasta_reader fasta_save, failed to write number sequence in output\n");
			perror("");
			exit(1);
		}
		current=(fasta*)current->next;
	}	
	
}

fasta * fasta_load(FILE * f) { 
	uint32_t length; 
	assert(sizeof(uint32_t)==4);
	size_t read = fread(&length,4,1,f);
	if (read!=1) {
		fprintf(stderr,"error: fasta_reader fasta_load, failed to read in number of sequences\n");
		perror("");
		exit(1);
	}
	length=ntohl(length); 
	if (length==0) {
		fprintf(stderr,"error: fasta_reader fasta_load, file contains no sequences\n");
		exit(1);
	}
	uint32_t lengths[length];
	read = fread(&lengths,4, length,f);
	if (read!=length) {
		fprintf(stderr,"error: fasta_reader fasta_load, failed to read in full binary header\n");
		perror("");
		exit(1);
	}
	int i;	
	for (i=0; i<length; i++) {
		lengths[i]=ntohl(lengths[i]);
	}
	//have read in header, now read in sequences
	//allocate and fill head
	fasta head; fasta * previous=&head;
	for (i=0; i<length; i++) {
		//get the next node in the list
		fasta * current = (fasta*)malloc(sizeof(fasta));
		if (current==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed malloc for head\n");
			perror("");
			exit(1);
		}
		memset(current,0,sizeof(fasta));
		//find out the length of the name field
		uint32_t name_length;
		read=fread(&name_length,4,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string length\n");
			perror("");
			exit(1);
		}
		name_length=ntohl(name_length);
		//get the name
		char * name = (char*)malloc(name_length);
		if (name==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to allocate memory for header string\n");
			perror("");
			exit(1);
		}
		read=fread(name,name_length,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string\n");
			perror("");
			exit(1);
		}
		current->name=name;
		//find out length of sequence
		uint32_t sequence_length;
		read=fread(&sequence_length,4,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in sequence length\n");
			perror("");
			exit(1);
		}
		sequence_length=ntohl(sequence_length);
		//get the sequence
		unsigned char * number_sequence = (unsigned char *)malloc(sequence_length);
		if (number_sequence==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to allocate memory for number sequence\n");
			perror("");
			exit(1);
		}
		read=fread(number_sequence,sequence_length,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string\n");
			perror("");
			exit(1);
		}
		current->size=sequence_length;
		current->original_size=sequence_length;
		current->start=1;
		current->end=current->size;
		current->number_sequence=number_sequence;
		previous->next=(struct fasta *)current;	
		previous=current;
	}
	
	return (fasta*)head.next;
}
fasta * fasta_load_region(FILE * f, char * contig_name, size_t start, size_t end) {
	uint32_t length; 
	assert(sizeof(uint32_t)==4);
	size_t read = fread(&length,4,1,f);
	if (read!=1) {
		fprintf(stderr,"error: fasta_reader fasta_load, failed to read in number of sequences\n");
		perror("");
		exit(1);
	}
	length=ntohl(length); 
	if (length==0) {
		fprintf(stderr,"error: fasta_reader fasta_load, file contains no sequences\n");
		exit(1);
	}
	uint32_t lengths[length];
	read = fread(&lengths,4, length,f);
	if (read!=length) {
		fprintf(stderr,"error: fasta_reader fasta_load, failed to read in full binary header\n");
		perror("");
		exit(1);
	}
	int i;	
	for (i=0; i<length; i++) {
		lengths[i]=ntohl(lengths[i]);
	}
	//have read in header, now read in sequences
	for (i=0; i<length; i++) {
		//get the next node in the list
		fasta * current = (fasta*)malloc(sizeof(fasta));
		if (current==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed malloc for head\n");
			perror("");
			exit(1);
		}
		memset(current,0,sizeof(fasta));
		//find out the length of the name field
		uint32_t name_length;
		read=fread(&name_length,4,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string length\n");
			perror("");
			exit(1);
		}
		name_length=ntohl(name_length);
		//get the name
		char * name = (char*)malloc(name_length);
		if (name==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to allocate memory for header string\n");
			perror("");
			exit(1);
		}
		read=fread(name,name_length,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string\n");
			perror("");
			exit(1);
		}
		current->name=name;
		//find out length of sequence
		uint32_t sequence_length;
		read=fread(&sequence_length,4,1,f);
		if (read!=1) {
			fprintf(stderr,"error: fasta_reader fasta_load, failed to read in sequence length\n");
			perror("");
			exit(1);
		}
		sequence_length=ntohl(sequence_length);
		if (contig_name==NULL || strcmp(name,contig_name)==0) {
			//get the sequence
			current->original_size=sequence_length;
			if (end==0 || end>sequence_length) {
				end=sequence_length;
			}
			if (start==0) {
				start=1;
			}
			if (start>end) {
				start=end;
			}
			fseek(f,start-1,SEEK_CUR);
			sequence_length=end-start+1;
			
			unsigned char * number_sequence = (unsigned char *)malloc(sequence_length);
			if (number_sequence==NULL) {
				fprintf(stderr,"error: fasta_reader fasta_load, failed to allocate memory for number sequence\n");
				perror("");
				exit(1);
			}
			read=fread(number_sequence,sequence_length,1,f);
			if (read!=1) {
				fprintf(stderr,"error: fasta_reader fasta_load, failed to read in header string\n");
				perror("");
				exit(1);
			}
			current->size=sequence_length;
			current->start=start;
			current->end=end;
			current->number_sequence=number_sequence;
			return current;
		} 
		fseek(f,sequence_length,SEEK_CUR);
	}
	
	return NULL;
}


void fasta_crop(fasta * fa, long long unsigned start, long long unsigned end) {
	if (start==0) {
		start=1;
	}
	if (end==0) {
		end=fa->size;
	}
	if (start>end) {
		start=end;
	}
	if (fa->end<start || fa->end<end) {
		fprintf(stderr,"Invalid fasta_crop call!\n");
		fprintf(stderr,"Fasta file has size %llu, start %llu, end %llu\n",fa->size,fa->start,fa->end);
		fprintf(stderr,"Trying to crop area %llu to %llu\n",start, end);
		exit(1);
	}	
	fa->size=end-start+1;
	fa->start=start;
	fa->end=end;
	if (fa->colour_sequence!=NULL) {
		unsigned char * temp = malloc(end-start+1);
		if (temp==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_crop, malloc failed\n");
			perror("");
			exit(1);
		}
		int i;
		for (i=0; i<end-start+1; i++) {
			temp[i]=fa->colour_sequence[i+start-1];
		}
		free(fa->colour_sequence);
		fa->colour_sequence=temp;
	}
	if (fa->number_sequence!=NULL) {
		unsigned char * temp = malloc(end-start+1);
		if (temp==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_crop, malloc failed\n");
			perror("");
			exit(1);
		}
		int i;
		for (i=0; i<end-start+1; i++) {
			temp[i]=fa->number_sequence[i+start-1];
		}
		free(fa->number_sequence);
		fa->number_sequence=temp;
	}
	if (fa->sequence!=NULL) {
		char * temp = malloc(end-start+1+1);
		if (temp==NULL) {
			fprintf(stderr,"error: fasta_reader fasta_crop, malloc failed\n");
			perror("");
			exit(1);
		}
		int i;
		for (i=0; i<end-start+1; i++) {
			temp[i]=fa->sequence[i+start-1];
		}
		temp[i]='\0';
		free(fa->sequence);
		fa->sequence=temp;
	}
}


void fasta_free(fasta * fa) {
	if (fa->sequence!=NULL) {
		free(fa->sequence);
	}
	if (fa->number_sequence!=NULL) {
		free(fa->number_sequence);
	}
	if (fa->colour_sequence!=NULL) {
		free(fa->colour_sequence);
	}
	if (fa->name!=NULL) {
		free(fa->name);
	}
	if (fa->next!=NULL) {
		fasta_free((fasta*)fa->next);
	}
	free(fa);
}


fasta * fasta_pop_and_free(fasta * head,char * name) {
	fasta * current=head; fasta * previous=NULL;
	while (current!=NULL) {
		if (strcmp(name,current->name)==0) {
			break;
		}
		previous=current;
		current=(fasta*)current->next;
	}
	if (current==NULL) {
		return NULL;
	}
	if (previous==NULL) {
		if (current->next!=NULL) {
			fasta_free((fasta*)current->next);
		}
	} else {
		previous->next=current->next;
		fasta_free(head);	
	}
	return current;
}
